import keras

from keras_hub.src.layers.modeling.reversible_embedding import (
    ReversibleEmbedding,
)
from keras_hub.src.models.t5.t5_layer_norm import T5LayerNorm
from keras_hub.src.models.t5.t5_transformer_layer import T5TransformerLayer


class T5Encoder(keras.Model):
    def __init__(
        self,
        vocabulary_size,
        num_layers,
        num_heads,
        hidden_dim,
        intermediate_dim,
        key_value_dim=None,
        dropout=0.1,
        activation="relu",
        use_gated_activation=True,
        layer_norm_epsilon=1e-06,
        tie_embedding_weights=True,
        dtype=None,
        **kwargs,
    ):
        # === Layers ===
        self.token_embedding = ReversibleEmbedding(
            input_dim=vocabulary_size,
            output_dim=hidden_dim,
            tie_weights=tie_embedding_weights,
            embeddings_initializer=keras.initializers.TruncatedNormal(1.0),
            dtype=dtype,
            name="token_embedding",
        )
        self.encoder_embedding_dropout = keras.layers.Dropout(
            dropout,
            dtype=dtype,
            name="encoder_embedding_dropout",
        )
        self.encoder_transformer_layers = []
        for i in range(num_layers):
            layer = T5TransformerLayer(
                is_decoder=False,
                hidden_dim=hidden_dim,
                intermediate_dim=intermediate_dim,
                key_value_dim=key_value_dim or hidden_dim // num_heads,
                dropout=dropout,
                activation=activation,
                layer_norm_epsilon=layer_norm_epsilon,
                num_heads=num_heads,
                use_gated_activation=use_gated_activation,
                use_relative_attention_bias=bool(i == 0),
                dtype=dtype,
                name=f"transformer_encoder_layer_{i}",
            )
            self.encoder_transformer_layers.append(layer)
        self.encoder_layer_norm = T5LayerNorm(
            epsilon=layer_norm_epsilon,
            dtype=dtype,
            name="encoder_output_layer_norm",
        )
        self.encoder_dropout = keras.layers.Dropout(
            dropout,
            dtype=dtype,
            name="encoder_output_dropout",
        )

        # === Functional Model ===
        encoder_token_id_input = keras.Input(
            shape=(None,), dtype="int32", name="token_ids"
        )
        encoder_padding_mask_input = keras.Input(
            shape=(None,), dtype="int32", name="padding_mask"
        )
        # Encoder.
        x = self.token_embedding(encoder_token_id_input)
        x = self.encoder_embedding_dropout(x)
        encoder_attention_mask = encoder_padding_mask_input[:, None, :]
        position_bias = None
        for transformer_layer in self.encoder_transformer_layers:
            output = transformer_layer(
                x,
                attention_mask=encoder_attention_mask,
                position_bias=position_bias,
                use_causal_mask=False,
            )
            if isinstance(output, tuple):
                x, position_bias = output
        x = self.encoder_layer_norm(x)
        x = self.encoder_dropout(x)
        sequence_output = x

        super().__init__(
            {
                "token_ids": encoder_token_id_input,
                "padding_mask": encoder_padding_mask_input,
            },
            outputs=sequence_output,
            **kwargs,
        )

        # === Config ===
        self.vocabulary_size = vocabulary_size
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.activation = keras.activations.get(activation)
        self.key_value_dim = key_value_dim
        self.dropout = dropout
        self.use_gated_activation = use_gated_activation
        self.layer_norm_epsilon = layer_norm_epsilon
        self.tie_embedding_weights = tie_embedding_weights

        if dtype is not None:
            try:
                self.dtype_policy = keras.dtype_policies.get(dtype)
            # Before Keras 3.2, there is no `keras.dtype_policies.get`.
            except AttributeError:
                if isinstance(dtype, keras.DTypePolicy):
                    dtype = dtype.name
                self.dtype_policy = keras.DTypePolicy(dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary_size": self.vocabulary_size,
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "num_layers": self.num_layers,
                "num_heads": self.num_heads,
                "activation": keras.activations.serialize(self.activation),
                "key_value_dim": self.key_value_dim,
                "dropout": self.dropout,
                "use_gated_activation": self.use_gated_activation,
                "layer_norm_epsilon": self.layer_norm_epsilon,
                "tie_embedding_weights": self.tie_embedding_weights,
            }
        )
        return config



"""StableDiffusion3 preset configurations."""

backbone_presets = {
    "stable_diffusion_3_medium": {
        "metadata": {
            "description": (
                "3 billion parameter, including CLIP L and CLIP G text "
                "encoders, MMDiT generative model, and VAE autoencoder. "
                "Developed by Stability AI."
            ),
            "params": 2987080931,
            "path": "stable_diffusion_3",
        },
        "kaggle_handle": "kaggle://keras/stablediffusion3/keras/stable_diffusion_3_medium/4",
    },
    "stable_diffusion_3.5_medium": {
        "metadata": {
            "description": (
                "3 billion parameter, including CLIP L and CLIP G text "
                "encoders, MMDiT-X generative model, and VAE autoencoder. "
                "Developed by Stability AI."
            ),
            "params": 3371793763,
            "path": "stable_diffusion_3",
        },
        "kaggle_handle": "kaggle://keras/stablediffusion3/keras/stable_diffusion_3.5_medium/1",
    },
    "stable_diffusion_3.5_large": {
        "metadata": {
            "description": (
                "9 billion parameter, including CLIP L and CLIP G text "
                "encoders, MMDiT generative model, and VAE autoencoder. "
                "Developed by Stability AI."
            ),
            "params": 9048410595,
            "path": "stable_diffusion_3",
        },
        "kaggle_handle": "kaggle://keras/stablediffusion-3.5/keras/stable_diffusion_3.5_large/2",
    },
    "stable_diffusion_3.5_large_turbo": {
        "metadata": {
            "description": (
                "9 billion parameter, including CLIP L and CLIP G text "
                "encoders, MMDiT generative model, and VAE autoencoder. "
                "A timestep-distilled version that eliminates classifier-free "
                "guidance and uses fewer steps for generation. "
                "Developed by Stability AI."
            ),
            "params": 9048410595,
            "path": "stable_diffusion_3",
        },
        "kaggle_handle": "kaggle://keras/stablediffusion-3.5/keras/stable_diffusion_3.5_large_turbo/2",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.basnet.basnet_backbone import BASNetBackbone
from keras_hub.src.models.basnet.basnet_image_converter import (
    BASNetImageConverter,
)
from keras_hub.src.models.image_segmenter_preprocessor import (
    ImageSegmenterPreprocessor,
)


@keras_hub_export("keras_hub.models.BASNetPreprocessor")
class BASNetPreprocessor(ImageSegmenterPreprocessor):
    backbone_cls = BASNetBackbone
    image_converter_cls = BASNetImageConverter



from keras_hub.src.models.mistral.mistral_backbone import MistralBackbone
from keras_hub.src.models.mistral.mistral_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, MistralBackbone)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_hub.src.models.albert.albert_tokenizer import AlbertTokenizer
from keras_hub.src.models.bart.bart_tokenizer import BartTokenizer
from keras_hub.src.models.bert.bert_tokenizer import BertTokenizer
from keras_hub.src.models.bloom.bloom_tokenizer import BloomTokenizer
from keras_hub.src.models.clip.clip_tokenizer import CLIPTokenizer
from keras_hub.src.models.deberta_v3.deberta_v3_tokenizer import (
    DebertaV3Tokenizer,
)
from keras_hub.src.models.distil_bert.distil_bert_tokenizer import (
    DistilBertTokenizer,
)
from keras_hub.src.models.electra.electra_tokenizer import ElectraTokenizer
from keras_hub.src.models.f_net.f_net_tokenizer import FNetTokenizer
from keras_hub.src.models.falcon.falcon_tokenizer import FalconTokenizer
from keras_hub.src.models.gemma.gemma_tokenizer import GemmaTokenizer
from keras_hub.src.models.gpt2.gpt2_tokenizer import GPT2Tokenizer
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_tokenizer import GPTNeoXTokenizer
from keras_hub.src.models.llama.llama_tokenizer import LlamaTokenizer
from keras_hub.src.models.llama3.llama3_tokenizer import Llama3Tokenizer
from keras_hub.src.models.mistral.mistral_tokenizer import MistralTokenizer
from keras_hub.src.models.opt.opt_tokenizer import OPTTokenizer
from keras_hub.src.models.pali_gemma.pali_gemma_tokenizer import (
    PaliGemmaTokenizer,
)
from keras_hub.src.models.phi3.phi3_tokenizer import Phi3Tokenizer
from keras_hub.src.models.roberta.roberta_tokenizer import RobertaTokenizer
from keras_hub.src.models.t5.t5_tokenizer import T5Tokenizer
from keras_hub.src.models.whisper.whisper_tokenizer import WhisperTokenizer
from keras_hub.src.models.xlm_roberta.xlm_roberta_tokenizer import (
    XLMRobertaTokenizer,
)
from keras_hub.src.tokenizers.byte_pair_tokenizer import BytePairTokenizer
from keras_hub.src.tokenizers.byte_tokenizer import ByteTokenizer
from keras_hub.src.tokenizers.sentence_piece_tokenizer import (
    SentencePieceTokenizer,
)
from keras_hub.src.tokenizers.sentence_piece_tokenizer_trainer import (
    compute_sentence_piece_proto,
)
from keras_hub.src.tokenizers.tokenizer import Tokenizer
from keras_hub.src.tokenizers.unicode_codepoint_tokenizer import (
    UnicodeCodepointTokenizer,
)
from keras_hub.src.tokenizers.word_piece_tokenizer import WordPieceTokenizer
from keras_hub.src.tokenizers.word_piece_tokenizer_trainer import (
    compute_word_piece_vocabulary,
)



from typing import Any

import numpy as np

from keras_hub.src.models.vgg.vgg_backbone import VGGBackbone
from keras_hub.src.models.vgg.vgg_image_classifier import VGGImageClassifier

backbone_cls = VGGBackbone


REPEATS_BY_SIZE = {
    "vgg11": [1, 1, 2, 2, 2],
    "vgg13": [2, 2, 2, 2, 2],
    "vgg16": [2, 2, 3, 3, 3],
    "vgg19": [2, 2, 4, 4, 4],
}


def convert_backbone_config(timm_config):
    architecture = timm_config["architecture"]
    stackwise_num_repeats = REPEATS_BY_SIZE[architecture]
    return dict(
        stackwise_num_repeats=stackwise_num_repeats,
        stackwise_num_filters=[64, 128, 256, 512, 512],
    )


def convert_conv2d(
    model,
    loader,
    keras_layer_name: str,
    hf_layer_name: str,
):
    loader.port_weight(
        model.get_layer(keras_layer_name).kernel,
        hf_weight_key=f"{hf_layer_name}.weight",
        hook_fn=lambda x, _: np.transpose(x, (2, 3, 1, 0)),
    )
    loader.port_weight(
        model.get_layer(keras_layer_name).bias,
        hf_weight_key=f"{hf_layer_name}.bias",
    )


def convert_weights(
    backbone: VGGBackbone,
    loader,
    timm_config: dict[Any],
):
    architecture = timm_config["architecture"]
    stackwise_num_repeats = REPEATS_BY_SIZE[architecture]

    hf_index_to_keras_layer_name = {}
    layer_index = 0
    for block_index, repeats_in_block in enumerate(stackwise_num_repeats):
        for repeat_index in range(repeats_in_block):
            hf_index = layer_index
            layer_index += 2  # Conv + activation layers.
            layer_name = f"block{block_index + 1}_conv{repeat_index + 1}"
            hf_index_to_keras_layer_name[hf_index] = layer_name
        layer_index += 1  # Pooling layer after blocks.

    for hf_index, keras_layer_name in hf_index_to_keras_layer_name.items():
        convert_conv2d(
            backbone, loader, keras_layer_name, f"features.{hf_index}"
        )


def convert_head(
    task: VGGImageClassifier,
    loader,
    timm_config: dict[Any],
):
    convert_conv2d(task.head, loader, "fc1", "pre_logits.fc1")
    convert_conv2d(task.head, loader, "fc2", "pre_logits.fc2")

    loader.port_weight(
        task.head.get_layer("predictions").kernel,
        hf_weight_key="head.fc.weight",
        hook_fn=lambda x, _: np.transpose(np.squeeze(x)),
    )
    loader.port_weight(
        task.head.get_layer("predictions").bias,
        hf_weight_key="head.fc.bias",
    )



from keras import dtype_policies
from keras import layers
from keras import ops


def quick_gelu(x):
    return x * ops.sigmoid(1.702 * x)


# TODO: Deprecate this in favor of `keras.layers.MultiHeadAttention` once the
# dtype compatibility issue is resolved.
class CLIPMultiHeadAttention(layers.MultiHeadAttention):
    def _masked_softmax(self, attention_scores, attention_mask=None):
        attention_scores = super()._masked_softmax(
            attention_scores, attention_mask
        )
        return ops.cast(attention_scores, self._value_dense.compute_dtype)


class CLIPEncoderBlock(layers.Layer):
    def __init__(
        self,
        hidden_dim,
        num_heads,
        intermediate_dim,
        intermediate_activation="quick_gelu",
        use_causal_mask=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if hidden_dim % num_heads != 0:
            raise ValueError(
                "`hidden_dim` must be divisible by `num_heads`. "
                f"Received: hidden_dim={hidden_dim}, num_heads={num_heads}"
            )
        self.hidden_dim = hidden_dim
        self.num_heads = num_heads
        self.intermediate_dim = intermediate_dim
        self.intermediate_activation = intermediate_activation
        self.use_causal_mask = use_causal_mask

        if intermediate_activation == "quick_gelu":
            intermediate_activation = quick_gelu

        self.layer_norm_1 = layers.LayerNormalization(
            epsilon=1e-5, dtype=self.dtype_policy, name="layer_norm_1"
        )
        self.attention = CLIPMultiHeadAttention(
            num_heads,
            hidden_dim // num_heads,
            dtype=self.dtype_policy,
            name="attention",
        )
        self.layer_norm_2 = layers.LayerNormalization(
            epsilon=1e-5, dtype=self.dtype_policy, name="layer_norm_2"
        )
        self.dense_1 = layers.Dense(
            self.intermediate_dim, dtype=self.dtype_policy, name="dense_1"
        )
        self.activation = layers.Activation(
            intermediate_activation, dtype=self.dtype_policy, name="activation"
        )
        self.dense_2 = layers.Dense(
            self.hidden_dim, dtype=self.dtype_policy, name="dense_2"
        )

    def build(self, input_shape):
        self.layer_norm_1.build(input_shape)
        self.attention.build(input_shape, input_shape, input_shape)
        # Before Keras 3.2, there was no setter for `dtype_policy`. Directly
        # assign a `DTypePolicy` instead.
        self.attention._softmax.dtype_policy = dtype_policies.DTypePolicy(
            "float32"
        )
        self.layer_norm_2.build(input_shape)
        self.dense_1.build(input_shape)
        input_shape = self.dense_1.compute_output_shape(input_shape)
        self.dense_2.build(input_shape)

    def compute_output_shape(self, inputs_shape):
        outputs_shape = list(inputs_shape)
        outputs_shape[-1] = self.hidden_dim
        return outputs_shape

    def call(self, x, training=None):
        residual = x
        x = self.layer_norm_1(x)
        x = self.attention(
            x, x, x, training=training, use_causal_mask=self.use_causal_mask
        )
        x = ops.add(residual, x)

        residual = x
        x = self.dense_1(self.layer_norm_2(residual))
        x = self.activation(x)
        x = self.dense_2(x)
        x = ops.add(residual, x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "num_heads": self.num_heads,
                "intermediate_dim": self.intermediate_dim,
                "intermediate_activation": self.intermediate_activation,
                "use_causal_mask": self.use_causal_mask,
            }
        )
        return config



import keras
from keras import layers
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.stable_diffusion_3.flow_match_euler_discrete_scheduler import (  # noqa: E501
    FlowMatchEulerDiscreteScheduler,
)
from keras_hub.src.models.stable_diffusion_3.mmdit import MMDiT
from keras_hub.src.utils.keras_utils import standardize_data_format


class CLIPProjection(layers.Layer):
    def __init__(self, hidden_dim, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)

        self.dense = layers.Dense(
            hidden_dim,
            use_bias=False,
            dtype=self.dtype_policy,
            name="dense",
        )

    def build(self, inputs_shape, token_ids_shape):
        inputs_shape = list(inputs_shape)
        self.dense.build([None, inputs_shape[-1]])

        # Assign identity matrix to the kernel as default.
        self.dense._kernel.assign(ops.eye(self.hidden_dim))

    def call(self, inputs, token_ids):
        indices = ops.expand_dims(
            ops.cast(ops.argmax(token_ids, axis=-1), "int32"), axis=-1
        )
        pooled_output = ops.take_along_axis(inputs, indices[:, :, None], axis=1)
        pooled_output = ops.squeeze(pooled_output, axis=1)
        return self.dense(pooled_output)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        return (inputs_shape[0], self.hidden_dim)


class CLIPConcatenate(layers.Layer):
    def call(
        self,
        clip_l_projection,
        clip_g_projection,
        clip_l_intermediate_output,
        clip_g_intermediate_output,
        padding,
    ):
        pooled_embeddings = ops.concatenate(
            [clip_l_projection, clip_g_projection], axis=-1
        )
        embeddings = ops.concatenate(
            [clip_l_intermediate_output, clip_g_intermediate_output], axis=-1
        )
        embeddings = ops.pad(embeddings, [[0, 0], [0, 0], [0, padding]])
        return pooled_embeddings, embeddings


class ImageRescaling(layers.Rescaling):
    """Rescales inputs from image space to latent space.

    The rescaling is performed using the formula: `(inputs - offset) * scale`.
    """

    def call(self, inputs):
        dtype = self.compute_dtype
        scale = self.backend.cast(self.scale, dtype)
        offset = self.backend.cast(self.offset, dtype)
        return (self.backend.cast(inputs, dtype) - offset) * scale


class LatentRescaling(layers.Rescaling):
    """Rescales inputs from latent space to image space.

    The rescaling is performed using the formula: `inputs / scale + offset`.
    """

    def call(self, inputs):
        dtype = self.compute_dtype
        scale = self.backend.cast(self.scale, dtype)
        offset = self.backend.cast(self.offset, dtype)
        return (self.backend.cast(inputs, dtype) / scale) + offset


class ClassifierFreeGuidanceConcatenate(layers.Layer):
    def call(
        self,
        latents,
        positive_contexts,
        negative_contexts,
        positive_pooled_projections,
        negative_pooled_projections,
        timestep,
    ):
        timestep = ops.broadcast_to(timestep, ops.shape(latents)[:1])
        latents = ops.concatenate([latents, latents], axis=0)
        contexts = ops.concatenate(
            [positive_contexts, negative_contexts], axis=0
        )
        pooled_projections = ops.concatenate(
            [positive_pooled_projections, negative_pooled_projections], axis=0
        )
        timesteps = ops.concatenate([timestep, timestep], axis=0)
        return latents, contexts, pooled_projections, timesteps


class ClassifierFreeGuidance(layers.Layer):
    """Perform classifier free guidance.

    This layer expects the inputs to be a concatenation of positive and negative
    (or empty) noise. The computation applies the classifier-free guidance
    scale.

    Args:
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.

    Call arguments:
        inputs: A concatenation of positive and negative (or empty) noises.
        guidance_scale: The scale factor for classifier-free guidance.

    Reference:
    - [Classifier-Free Diffusion Guidance](https://arxiv.org/abs/2207.12598).
    """

    def call(self, inputs, guidance_scale):
        positive_noise, negative_noise = ops.split(inputs, 2, axis=0)
        return ops.add(
            negative_noise,
            ops.multiply(
                guidance_scale, ops.subtract(positive_noise, negative_noise)
            ),
        )

    def compute_output_shape(self, inputs_shape):
        outputs_shape = list(inputs_shape)
        if outputs_shape[0] is not None:
            outputs_shape[0] = outputs_shape[0] // 2
        return outputs_shape


class EulerStep(layers.Layer):
    """A layer predicts the sample with the timestep and the predicted noise.

    Args:
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.

    Call arguments:
        latents: A current sample created by the diffusion process.
        noise_residual: The direct output from the diffusion model.
        sigma: The amount of noise added at the current timestep.
        sigma_next: The amount of noise added at the next timestep.

    References:
    - [Common Diffusion Noise Schedules and Sample Steps are Flawed](
    https://arxiv.org/abs/2305.08891).
    - [Elucidating the Design Space of Diffusion-Based Generative Models](
    https://arxiv.org/abs/2206.00364).
    """

    def call(self, latents, noise_residual, sigma, sigma_next):
        sigma_diff = ops.subtract(sigma_next, sigma)
        return ops.add(latents, ops.multiply(sigma_diff, noise_residual))

    def compute_output_shape(self, latents_shape):
        return latents_shape


@keras_hub_export("keras_hub.models.StableDiffusion3Backbone")
class StableDiffusion3Backbone(Backbone):
    """Stable Diffusion 3 core network with hyperparameters.

    This backbone imports CLIP and T5 models as text encoders and implements the
    base MMDiT and VAE networks for the Stable Diffusion 3 model.

    The default constructor gives a fully customizable, randomly initialized
    MMDiT and VAE models with any hyperparameters. To load preset architectures
    and weights, use the `from_preset` constructor.

    Args:
        mmdit_patch_size: int. The size of each square patch in the input image
            in MMDiT.
        mmdit_hidden_dim: int. The size of the transformer hidden state at the
            end of each transformer layer in MMDiT.
        mmdit_num_layers: int. The number of transformer layers in MMDiT.
        mmdit_num_heads: int. The number of attention heads for each
            transformer in MMDiT.
        mmdit_position_size: int. The size of the height and width for the
            position embedding in MMDiT.
        mmdit_qk_norm: Optional str. Whether to normalize the query and key
            tensors for each transformer in MMDiT. Available options are `None`
            and `"rms_norm"`. Typically, this is set to `None` for 3.0 version
            and to `"rms_norm"` for 3.5 version.
        mmdit_dual_attention_indices: Optional tuple. Specifies the indices of
            the blocks that serve as dual attention blocks. Typically, this is
            for 3.5 version. Defaults to `None`.
        vae: The VAE used for transformations between pixel space and latent
            space.
        clip_l: The CLIP text encoder for encoding the inputs.
        clip_g: The CLIP text encoder for encoding the inputs.
        t5: optional The T5 text encoder for encoding the inputs.
        latent_channels: int. The number of channels in the latent. Defaults to
            `16`.
        output_channels: int. The number of channels in the output. Defaults to
            `3`.
        num_train_timesteps: int. The number of diffusion steps to train the
            model. Defaults to `1000`.
        shift: float. The shift value for the timestep schedule. Defaults to
            `3.0`.
        image_shape: tuple. The input shape without the batch size. Defaults to
            `(1024, 1024, 3)`.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        dtype: string or `keras.mixed_precision.DTypePolicy`. The dtype to use
            for the models computations and weights. Note that some
            computations, such as softmax and layer normalization will always
            be done a float32 precision regardless of dtype.

    Example:
    ```python
    # Pretrained Stable Diffusion 3 model.
    model = keras_hub.models.StableDiffusion3Backbone.from_preset(
        "stable_diffusion_3_medium"
    )

    # Randomly initialized Stable Diffusion 3 model with custom config.
    vae = keras_hub.models.VAEBackbone(...)
    clip_l = keras_hub.models.CLIPTextEncoder(...)
    clip_g = keras_hub.models.CLIPTextEncoder(...)
    model = keras_hub.models.StableDiffusion3Backbone(
        mmdit_patch_size=2,
        mmdit_num_heads=4,
        mmdit_hidden_dim=256,
        mmdit_depth=4,
        mmdit_position_size=192,
        mmdit_qk_norm=None,
        mmdit_dual_attention_indices=None,
        vae=vae,
        clip_l=clip_l,
        clip_g=clip_g,
    )
    ```
    """

    def __init__(
        self,
        mmdit_patch_size,
        mmdit_hidden_dim,
        mmdit_num_layers,
        mmdit_num_heads,
        mmdit_position_size,
        mmdit_qk_norm,
        mmdit_dual_attention_indices,
        vae,
        clip_l,
        clip_g,
        t5=None,
        latent_channels=16,
        output_channels=3,
        num_train_timesteps=1000,
        shift=3.0,
        image_shape=(1024, 1024, 3),
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        data_format = standardize_data_format(data_format)
        if data_format != "channels_last":
            raise NotImplementedError
        height = image_shape[0]
        width = image_shape[1]
        if height % 8 != 0 or width % 8 != 0:
            raise ValueError(
                "height and width in `image_shape` must be divisible by 8. "
                f"Received: image_shape={image_shape}"
            )
        latent_shape = (height // 8, width // 8, int(latent_channels))
        context_shape = (None, 4096 if t5 is None else t5.hidden_dim)
        pooled_projection_shape = (clip_l.hidden_dim + clip_g.hidden_dim,)
        self._latent_shape = latent_shape

        # === Layers ===
        self.clip_l = clip_l
        self.clip_l_projection = CLIPProjection(
            clip_l.hidden_dim, dtype=dtype, name="clip_l_projection"
        )
        self.clip_g = clip_g
        self.clip_g_projection = CLIPProjection(
            clip_g.hidden_dim, dtype=dtype, name="clip_g_projection"
        )
        self.clip_concatenate = CLIPConcatenate(
            dtype=dtype, name="clip_concatenate"
        )
        self.t5 = t5
        self.diffuser = MMDiT(
            mmdit_patch_size,
            mmdit_hidden_dim,
            mmdit_num_layers,
            mmdit_num_heads,
            mmdit_position_size,
            latent_shape=latent_shape,
            context_shape=context_shape,
            pooled_projection_shape=pooled_projection_shape,
            qk_norm=mmdit_qk_norm,
            dual_attention_indices=mmdit_dual_attention_indices,
            data_format=data_format,
            dtype=dtype,
            name="diffuser",
        )
        self.vae = vae
        self.cfg_concat = ClassifierFreeGuidanceConcatenate(
            dtype=dtype, name="classifier_free_guidance_concat"
        )
        self.cfg = ClassifierFreeGuidance(
            dtype=dtype, name="classifier_free_guidance"
        )
        # Set `dtype="float32"` to ensure the high precision for the noise
        # residual.
        self.scheduler = FlowMatchEulerDiscreteScheduler(
            num_train_timesteps=num_train_timesteps,
            shift=shift,
            dtype="float32",
            name="scheduler",
        )
        self.euler_step = EulerStep(dtype="float32", name="euler_step")
        self.image_rescaling = ImageRescaling(
            scale=self.vae.scale,
            offset=self.vae.shift,
            dtype=dtype,
            name="image_rescaling",
        )
        self.latent_rescaling = LatentRescaling(
            scale=self.vae.scale,
            offset=self.vae.shift,
            dtype=dtype,
            name="latent_rescaling",
        )

        # === Functional Model ===
        image_input = keras.Input(
            shape=image_shape,
            name="images",
        )
        latent_input = keras.Input(
            shape=latent_shape,
            name="latents",
        )
        clip_l_token_id_input = keras.Input(
            shape=(None,),
            dtype="int32",
            name="clip_l_token_ids",
        )
        clip_l_negative_token_id_input = keras.Input(
            shape=(None,),
            dtype="int32",
            name="clip_l_negative_token_ids",
        )
        clip_g_token_id_input = keras.Input(
            shape=(None,),
            dtype="int32",
            name="clip_g_token_ids",
        )
        clip_g_negative_token_id_input = keras.Input(
            shape=(None,),
            dtype="int32",
            name="clip_g_negative_token_ids",
        )
        token_ids = {
            "clip_l": clip_l_token_id_input,
            "clip_g": clip_g_token_id_input,
        }
        negative_token_ids = {
            "clip_l": clip_l_negative_token_id_input,
            "clip_g": clip_g_negative_token_id_input,
        }
        if self.t5 is not None:
            t5_token_id_input = keras.Input(
                shape=(None,),
                dtype="int32",
                name="t5_token_ids",
            )
            t5_negative_token_id_input = keras.Input(
                shape=(None,),
                dtype="int32",
                name="t5_negative_token_ids",
            )
            token_ids["t5"] = t5_token_id_input
            negative_token_ids["t5"] = t5_negative_token_id_input
        num_step_input = keras.Input(
            shape=(),
            dtype="int32",
            name="num_steps",
        )
        guidance_scale_input = keras.Input(
            shape=(),
            dtype="float32",
            name="guidance_scale",
        )
        embeddings = self.encode_text_step(token_ids, negative_token_ids)
        latents = self.encode_image_step(image_input)
        # Use `steps=0` to define the functional model.
        denoised_latents = self.denoise_step(
            latent_input,
            embeddings,
            0,
            num_step_input[0],
            guidance_scale_input[0],
        )
        images = self.decode_step(denoised_latents)
        inputs = {
            "images": image_input,
            "latents": latent_input,
            "clip_l_token_ids": clip_l_token_id_input,
            "clip_l_negative_token_ids": clip_l_negative_token_id_input,
            "clip_g_token_ids": clip_g_token_id_input,
            "clip_g_negative_token_ids": clip_g_negative_token_id_input,
            "num_steps": num_step_input,
            "guidance_scale": guidance_scale_input,
        }
        outputs = {
            "latents": latents,
            "images": images,
        }
        if self.t5 is not None:
            inputs["t5_token_ids"] = t5_token_id_input
            inputs["t5_negative_token_ids"] = t5_negative_token_id_input
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            dtype=dtype,
            **kwargs,
        )

        # === Config ===
        self.mmdit_patch_size = mmdit_patch_size
        self.mmdit_hidden_dim = mmdit_hidden_dim
        self.mmdit_num_layers = mmdit_num_layers
        self.mmdit_num_heads = mmdit_num_heads
        self.mmdit_position_size = mmdit_position_size
        self.mmdit_qk_norm = mmdit_qk_norm
        self.mmdit_dual_attention_indices = mmdit_dual_attention_indices
        self.latent_channels = latent_channels
        self.output_channels = output_channels
        self.num_train_timesteps = num_train_timesteps
        self.shift = shift
        self.image_shape = image_shape

    @property
    def latent_shape(self):
        return (None,) + self._latent_shape

    @property
    def clip_hidden_dim(self):
        return self.clip_l.hidden_dim + self.clip_g.hidden_dim

    @property
    def t5_hidden_dim(self):
        return 4096 if self.t5 is None else self.t5.hidden_dim

    def encode_text_step(self, token_ids, negative_token_ids):
        clip_hidden_dim = self.clip_hidden_dim
        t5_hidden_dim = self.t5_hidden_dim

        def encode(token_ids):
            clip_l_outputs = self.clip_l(
                {"token_ids": token_ids["clip_l"]}, training=False
            )
            clip_g_outputs = self.clip_g(
                {"token_ids": token_ids["clip_g"]}, training=False
            )
            clip_l_projection = self.clip_l_projection(
                clip_l_outputs["sequence_output"],
                token_ids["clip_l"],
                training=False,
            )
            clip_g_projection = self.clip_g_projection(
                clip_g_outputs["sequence_output"],
                token_ids["clip_g"],
                training=False,
            )
            pooled_embeddings, embeddings = self.clip_concatenate(
                clip_l_projection,
                clip_g_projection,
                clip_l_outputs["intermediate_output"],
                clip_g_outputs["intermediate_output"],
                padding=t5_hidden_dim - clip_hidden_dim,
            )
            if self.t5 is not None:
                t5_outputs = self.t5(
                    {
                        "token_ids": token_ids["t5"],
                        "padding_mask": ops.ones_like(token_ids["t5"]),
                    },
                    training=False,
                )
                embeddings = ops.concatenate([embeddings, t5_outputs], axis=-2)
            else:
                padded_size = self.clip_l.max_sequence_length
                embeddings = ops.pad(
                    embeddings, [[0, 0], [0, padded_size], [0, 0]]
                )
            return embeddings, pooled_embeddings

        positive_embeddings, positive_pooled_embeddings = encode(token_ids)
        negative_embeddings, negative_pooled_embeddings = encode(
            negative_token_ids
        )
        return (
            positive_embeddings,
            negative_embeddings,
            positive_pooled_embeddings,
            negative_pooled_embeddings,
        )

    def encode_image_step(self, images):
        latents = self.vae.encode(images)
        return self.image_rescaling(latents)

    def add_noise_step(self, latents, noises, step, num_steps):
        return self.scheduler.add_noise(latents, noises, step, num_steps)

    def denoise_step(
        self,
        latents,
        embeddings,
        step,
        num_steps,
        guidance_scale=None,
    ):
        step = ops.convert_to_tensor(step)
        next_step = ops.add(step, 1)
        sigma, timestep = self.scheduler(step, num_steps)
        next_sigma, _ = self.scheduler(next_step, num_steps)

        # Concatenation for classifier-free guidance.
        if guidance_scale is not None:
            concated_latents, contexts, pooled_projs, timesteps = (
                self.cfg_concat(latents, *embeddings, timestep)
            )
        else:
            timesteps = ops.broadcast_to(timestep, ops.shape(latents)[:1])
            concated_latents = latents
            contexts = embeddings[0]
            pooled_projs = embeddings[2]

        # Diffusion.
        predicted_noise = self.diffuser(
            {
                "latent": concated_latents,
                "context": contexts,
                "pooled_projection": pooled_projs,
                "timestep": timesteps,
            },
            training=False,
        )

        # Classifier-free guidance.
        if guidance_scale is not None:
            predicted_noise = self.cfg(predicted_noise, guidance_scale)

        # Euler step.
        return self.euler_step(latents, predicted_noise, sigma, next_sigma)

    def decode_step(self, latents):
        latents = self.latent_rescaling(latents)
        return self.vae.decode(latents, training=False)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "mmdit_patch_size": self.mmdit_patch_size,
                "mmdit_hidden_dim": self.mmdit_hidden_dim,
                "mmdit_num_layers": self.mmdit_num_layers,
                "mmdit_num_heads": self.mmdit_num_heads,
                "mmdit_position_size": self.mmdit_position_size,
                "mmdit_qk_norm": self.mmdit_qk_norm,
                "mmdit_dual_attention_indices": (
                    self.mmdit_dual_attention_indices
                ),
                "vae": layers.serialize(self.vae),
                "clip_l": layers.serialize(self.clip_l),
                "clip_g": layers.serialize(self.clip_g),
                "t5": layers.serialize(self.t5),
                "latent_channels": self.latent_channels,
                "output_channels": self.output_channels,
                "num_train_timesteps": self.num_train_timesteps,
                "shift": self.shift,
                "image_shape": self.image_shape,
            }
        )
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = config.copy()

        # Propagate `dtype` to text encoders if needed.
        if "dtype" in config and config["dtype"] is not None:
            dtype_config = config["dtype"]
            if "dtype" not in config["vae"]["config"]:
                config["vae"]["config"]["dtype"] = dtype_config
            if "dtype" not in config["clip_l"]["config"]:
                config["clip_l"]["config"]["dtype"] = dtype_config
            if "dtype" not in config["clip_g"]["config"]:
                config["clip_g"]["config"]["dtype"] = dtype_config
            if (
                config["t5"] is not None
                and "dtype" not in config["t5"]["config"]
            ):
                config["t5"]["config"]["dtype"] = dtype_config

        # We expect `vae`, `clip_l`, `clip_g` and/or `t5` to be instantiated.
        config["vae"] = layers.deserialize(
            config["vae"], custom_objects=custom_objects
        )
        config["clip_l"] = layers.deserialize(
            config["clip_l"], custom_objects=custom_objects
        )
        config["clip_g"] = layers.deserialize(
            config["clip_g"], custom_objects=custom_objects
        )
        if config["t5"] is not None:
            config["t5"] = layers.deserialize(
                config["t5"], custom_objects=custom_objects
            )

        # To maintain backward compatibility, we need to ensure that
        # `mmdit_qk_norm` and `mmdit_dual_attention_indices` is included in the
        # config.
        if "mmdit_qk_norm" not in config:
            config["mmdit_qk_norm"] = None
        if "mmdit_dual_attention_indices" not in config:
            config["mmdit_dual_attention_indices"] = None
        return cls(**config)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)
from keras_hub.src.models.resnet.resnet_backbone import ResNetBackbone
from keras_hub.src.models.resnet.resnet_image_converter import (
    ResNetImageConverter,
)


@keras_hub_export("keras_hub.models.ResNetImageClassifierPreprocessor")
class ResNetImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = ResNetBackbone
    image_converter_cls = ResNetImageConverter



"""DeepLabV3 preset configurations."""

backbone_presets = {
    "deeplab_v3_plus_resnet50_pascalvoc": {
        "metadata": {
            "description": (
                "DeepLabV3+ model with ResNet50 as image encoder and trained "
                "on augmented Pascal VOC dataset by Semantic Boundaries "
                "Dataset(SBD) which is having categorical accuracy of 90.01 "
                "and 0.63 Mean IoU."
            ),
            "params": 39190656,
            "path": "deeplab_v3",
        },
        "kaggle_handle": "kaggle://keras/deeplabv3plus/keras/deeplab_v3_plus_resnet50_pascalvoc/4",
    },
}



from keras import layers
from keras import ops

from keras_hub.src.utils.keras_utils import standardize_data_format


class CLIPVisionEmbedding(layers.Layer):
    def __init__(
        self,
        hidden_dim,
        patch_size,
        image_size,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.hidden_dim = int(hidden_dim)
        self.patch_size = int(patch_size)
        self.image_size = int(image_size)
        data_format = standardize_data_format(data_format)
        self.data_format = data_format
        num_patches = (image_size // patch_size) ** 2
        self.num_positions = num_patches + 1

        self.patch_embedding = layers.Conv2D(
            hidden_dim,
            kernel_size=patch_size,
            strides=patch_size,
            data_format=data_format,
            use_bias=False,
            dtype=dtype,
            name="patch_embedding",
        )
        self.position_embedding = layers.Embedding(
            num_patches + 1, hidden_dim, dtype=dtype, name="position_embedding"
        )

    def build(self, input_shape):
        self.class_embedding = self.add_weight(
            shape=(self.hidden_dim,),
            initializer="random_normal",
            dtype=self.variable_dtype,
            name="class_embedding",
        )
        self.position_ids = self.add_weight(
            shape=(1, self.num_positions),
            initializer="zeros",
            # Let the backend determine the int dtype. For example, tf
            # requires int64 for correct device placement, whereas jax and torch
            # don't.
            dtype=int,
            trainable=False,
            name="position_ids",
        )
        self.patch_embedding.build(input_shape)
        self.position_embedding.build(self.position_ids.shape)

    def call(self, inputs, training=None):
        x = inputs
        batch_size = ops.shape(x)[0]
        patch_embeddings = self.patch_embedding(x, training=training)
        if self.data_format == "channels_last":
            patch_embeddings = ops.reshape(
                patch_embeddings, (batch_size, -1, self.hidden_dim)
            )
        else:
            patch_embeddings = ops.reshape(
                patch_embeddings, (batch_size, self.hidden_dim, -1)
            )
            patch_embeddings = ops.transpose(patch_embeddings, (0, 2, 1))
        class_embeddings = ops.expand_dims(self.class_embedding, axis=(0, 1))
        class_embeddings = ops.tile(class_embeddings, (batch_size, 1, 1))
        position_embeddings = self.position_embedding(self.position_ids)
        embeddings = ops.concatenate(
            [class_embeddings, patch_embeddings], axis=1
        )
        return ops.add(embeddings, position_embeddings)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "patch_size": self.patch_size,
                "image_size": self.image_size,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        output_shape = [input_shape[0], None, self.hidden_dim]
        if self.data_format == "channels_last":
            if input_shape[1] is not None and input_shape[2] is not None:
                patch_num = input_shape[1] // self.patch_size
                output_shape[1] = patch_num**2 + 1
        else:
            if input_shape[2] is not None and input_shape[3] is not None:
                patch_num = input_shape[2] // self.patch_size
                output_shape[1] = patch_num**2 + 1
        return output_shape



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier import ImageClassifier
from keras_hub.src.models.mobilenet.mobilenet_backbone import MobileNetBackbone
from keras_hub.src.models.mobilenet.mobilenet_image_classifier_preprocessor import (  # noqa: E501
    MobileNetImageClassifierPreprocessor,
)
from keras_hub.src.models.task import Task


@keras_hub_export("keras_hub.models.MobileNetImageClassifier")
class MobileNetImageClassifier(ImageClassifier):
    backbone_cls = MobileNetBackbone
    preprocessor_cls = MobileNetImageClassifierPreprocessor

    def __init__(
        self,
        backbone,
        num_classes,
        preprocessor=None,
        head_dtype=None,
        **kwargs,
    ):
        head_dtype = head_dtype or backbone.dtype_policy
        data_format = getattr(backbone, "data_format", None)

        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        self.pooler = keras.layers.GlobalAveragePooling2D(
            data_format, keepdims=True, dtype=head_dtype, name="pooler"
        )

        self.output_conv = keras.layers.Conv2D(
            filters=1024,
            kernel_size=(1, 1),
            strides=(1, 1),
            use_bias=True,
            padding="valid",
            activation="hard_silu",
            name="classifier_conv",
            dtype=head_dtype,
        )

        self.flatten = keras.layers.Flatten(
            dtype=head_dtype,
        )

        self.output_dense = keras.layers.Dense(
            num_classes,
            dtype=head_dtype,
            name="predictions",
        )

        # === Functional Model ===
        inputs = self.backbone.input
        x = self.backbone(inputs)
        x = self.pooler(x)
        x = self.output_conv(x)
        x = self.flatten(x)
        outputs = self.output_dense(x)
        Task.__init__(
            self,
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

        # === Config ===
        self.num_classes = num_classes

    def get_config(self):
        # Skip ImageClassifier
        config = Task.get_config(self)
        config.update(
            {
                "num_classes": self.num_classes,
            }
        )
        return config



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm_preprocessor import CausalLMPreprocessor
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_backbone import GPTNeoXBackbone
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_tokenizer import GPTNeoXTokenizer


@keras_hub_export("keras_hub.models.GPTNeoXCausalLMPreprocessor")
class GPTNeoXCausalLMPreprocessor(CausalLMPreprocessor):
    """GPT-NeoX Causal LM preprocessor.

    This preprocessing layer is meant for use with
    `keras_hub.models.GPTNeoXCausalLM`. By default, it will take in batches of
    strings, and return outputs in a `(x, y, sample_weight)` format, where the
    `y` label is the next token id in the `x` sequence.

    For use with generation, the layer also exposes two methods
    `generate_preprocess()` and `generate_postprocess()`. When this preprocessor
    is attached to a `keras_hub.models.GPTNeoXCausalLM` instance, these methods
    will be called implicitly in `generate()`. They can also be called
    standalone (e.g. to precompute preprocessing inputs for generation in a
    separate process).

    Args:
        tokenizer: A `keras_hub.models.GPTNeoXTokenizer` instance.
        sequence_length: The length of the packed inputs.
        add_start_token: If `True`, the preprocessor will prepend the tokenizer
            start token to each input sequence.
        add_end_token: If `True`, the preprocessor will append the tokenizer
            end token to each input sequence.

    Call arguments:
        x: A string, `tf.Tensor` or list of python strings.
        y: Label data. Should always be `None` as the layer generates labels.
        sample_weight: Label weights. Should always be `None` as the layer
            generates label weights.
        sequence_length: Pass to override the configured `sequence_length` of
            the layer.

    """

    backbone_cls = GPTNeoXBackbone
    tokenizer_cls = GPTNeoXTokenizer



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_hub.src.layers.modeling.alibi_bias import AlibiBias
from keras_hub.src.layers.modeling.anchor_generator import AnchorGenerator
from keras_hub.src.layers.modeling.box_matcher import BoxMatcher
from keras_hub.src.layers.modeling.cached_multi_head_attention import (
    CachedMultiHeadAttention,
)
from keras_hub.src.layers.modeling.f_net_encoder import FNetEncoder
from keras_hub.src.layers.modeling.masked_lm_head import MaskedLMHead
from keras_hub.src.layers.modeling.non_max_supression import NonMaxSuppression
from keras_hub.src.layers.modeling.position_embedding import PositionEmbedding
from keras_hub.src.layers.modeling.reversible_embedding import (
    ReversibleEmbedding,
)
from keras_hub.src.layers.modeling.rms_normalization import RMSNormalization
from keras_hub.src.layers.modeling.rotary_embedding import RotaryEmbedding
from keras_hub.src.layers.modeling.sine_position_encoding import (
    SinePositionEncoding,
)
from keras_hub.src.layers.modeling.token_and_position_embedding import (
    TokenAndPositionEmbedding,
)
from keras_hub.src.layers.modeling.transformer_decoder import TransformerDecoder
from keras_hub.src.layers.modeling.transformer_encoder import TransformerEncoder
from keras_hub.src.layers.preprocessing.audio_converter import AudioConverter
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.layers.preprocessing.masked_lm_mask_generator import (
    MaskedLMMaskGenerator,
)
from keras_hub.src.layers.preprocessing.multi_segment_packer import (
    MultiSegmentPacker,
)
from keras_hub.src.layers.preprocessing.random_deletion import RandomDeletion
from keras_hub.src.layers.preprocessing.random_swap import RandomSwap
from keras_hub.src.layers.preprocessing.start_end_packer import StartEndPacker
from keras_hub.src.models.basnet.basnet_image_converter import (
    BASNetImageConverter,
)
from keras_hub.src.models.clip.clip_image_converter import CLIPImageConverter
from keras_hub.src.models.deeplab_v3.deeplab_v3_image_converter import (
    DeepLabV3ImageConverter,
)
from keras_hub.src.models.densenet.densenet_image_converter import (
    DenseNetImageConverter,
)
from keras_hub.src.models.efficientnet.efficientnet_image_converter import (
    EfficientNetImageConverter,
)
from keras_hub.src.models.mit.mit_image_converter import MiTImageConverter
from keras_hub.src.models.mobilenet.mobilenet_image_converter import (
    MobileNetImageConverter,
)
from keras_hub.src.models.pali_gemma.pali_gemma_image_converter import (
    PaliGemmaImageConverter,
)
from keras_hub.src.models.resnet.resnet_image_converter import (
    ResNetImageConverter,
)
from keras_hub.src.models.retinanet.retinanet_image_converter import (
    RetinaNetImageConverter,
)
from keras_hub.src.models.sam.sam_image_converter import SAMImageConverter
from keras_hub.src.models.sam.sam_mask_decoder import SAMMaskDecoder
from keras_hub.src.models.sam.sam_prompt_encoder import SAMPromptEncoder
from keras_hub.src.models.segformer.segformer_image_converter import (
    SegFormerImageConverter,
)
from keras_hub.src.models.vgg.vgg_image_converter import VGGImageConverter
from keras_hub.src.models.vit.vit_image_converter import ViTImageConverter
from keras_hub.src.models.whisper.whisper_audio_converter import (
    WhisperAudioConverter,
)



import keras

BN_AXIS = 3


class CBABlock(keras.layers.Layer):
    """
    Args:
        input_filters: int, the number of input filters
        output_filters: int, the number of output filters
        kernel_size: default 3, the kernel_size to apply to the expansion phase
            convolutions
        strides: default 1, the strides to apply to the expansion phase
            convolutions
        data_format: str, channels_last (default) or channels_first, expects
            tensors to be of shape (N, H, W, C) or (N, C, H, W) respectively
        batch_norm_momentum: default 0.9, the BatchNormalization momentum
        batch_norm_epsilon: default 1e-3, the BatchNormalization epsilon
        activation: default "swish", the activation function used between
            convolution operations
        dropout: float, the optional dropout rate to apply before the output
            convolution, defaults to 0.2
        nores: bool, default False, forces no residual connection if True,
            otherwise allows it if False.

    Returns:
        A tensor representing a feature map, passed through the ConvBNAct
        block

    Note:
        Not intended to be used outside of the EfficientNet architecture.
    """

    def __init__(
        self,
        input_filters,
        output_filters,
        kernel_size=3,
        strides=1,
        data_format="channels_last",
        batch_norm_momentum=0.9,
        batch_norm_epsilon=1e-3,
        activation="swish",
        dropout=0.2,
        nores=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.input_filters = input_filters
        self.output_filters = output_filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.data_format = data_format
        self.batch_norm_momentum = batch_norm_momentum
        self.batch_norm_epsilon = batch_norm_epsilon
        self.activation = activation
        self.dropout = dropout
        self.nores = nores

        padding_pixels = kernel_size // 2
        self.conv1_pad = keras.layers.ZeroPadding2D(
            padding=(padding_pixels, padding_pixels),
            name=self.name + "conv_pad",
        )
        self.conv1 = keras.layers.Conv2D(
            filters=self.output_filters,
            kernel_size=kernel_size,
            strides=strides,
            kernel_initializer=self._conv_kernel_initializer(),
            padding="valid",
            data_format=data_format,
            use_bias=False,
            name=self.name + "conv",
        )
        self.bn1 = keras.layers.BatchNormalization(
            axis=BN_AXIS,
            momentum=self.batch_norm_momentum,
            epsilon=self.batch_norm_epsilon,
            name=self.name + "bn",
        )
        self.act = keras.layers.Activation(
            self.activation, name=self.name + "activation"
        )

        if self.dropout:
            self.dropout_layer = keras.layers.Dropout(
                self.dropout,
                noise_shape=(None, 1, 1, 1),
                name=self.name + "drop",
            )

    def _conv_kernel_initializer(
        self,
        scale=2.0,
        mode="fan_out",
        distribution="truncated_normal",
        seed=None,
    ):
        return keras.initializers.VarianceScaling(
            scale=scale, mode=mode, distribution=distribution, seed=seed
        )

    def build(self, input_shape):
        if self.name is None:
            self.name = keras.backend.get_uid("block0")

    def call(self, inputs):
        x = self.conv1_pad(inputs)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.act(x)

        # Residual:
        if (
            self.strides == 1
            and self.input_filters == self.output_filters
            and not self.nores
        ):
            if self.dropout:
                x = self.dropout_layer(x)
            x = keras.layers.Add(name=self.name + "add")([x, inputs])
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_filters": self.input_filters,
                "output_filters": self.output_filters,
                "kernel_size": self.kernel_size,
                "strides": self.strides,
                "data_format": self.data_format,
                "batch_norm_momentum": self.batch_norm_momentum,
                "batch_norm_epsilon": self.batch_norm_epsilon,
                "activation": self.activation,
                "dropout": self.dropout,
                "nores": self.nores,
            }
        )

        return config



"""FNet model preset configurations."""

backbone_presets = {
    "f_net_base_en": {
        "metadata": {
            "description": (
                "12-layer FNet model where case is maintained. "
                "Trained on the C4 dataset."
            ),
            "params": 82861056,
            "path": "f_net",
        },
        "kaggle_handle": "kaggle://keras/f_net/keras/f_net_base_en/3",
    },
    "f_net_large_en": {
        "metadata": {
            "description": (
                "24-layer FNet model where case is maintained. "
                "Trained on the C4 dataset."
            ),
            "params": 236945408,
            "path": "f_net",
        },
        "kaggle_handle": "kaggle://keras/f_net/keras/f_net_large_en/3",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.vit.vit_backbone import ViTBackbone
from keras_hub.src.utils.tensor_utils import preprocessing_function


@keras_hub_export("keras_hub.layers.ViTImageConverter")
class ViTImageConverter(ImageConverter):
    """Converts images to the format expected by a ViT model.

    This layer performs image normalization using mean and standard deviation
    values. By default, it uses the same normalization as the
    "google/vit-large-patch16-224" model on Hugging Face:
    `norm_mean=[0.5, 0.5, 0.5]` and `norm_std=[0.5, 0.5, 0.5]`
    ([reference](https://huggingface.co/google/vit-large-patch16-224/blob/main/preprocessor_config.json)).
    These defaults are suitable for models pretrained using this normalization.

    Args:
        norm_mean: list or tuple of floats. Mean values for image normalization.
            Defaults to `[0.5, 0.5, 0.5]`.
        norm_std: list or tuple of floats. Standard deviation values for
            image normalization. Defaults to `[0.5, 0.5, 0.5]`.
        **kwargs: Additional keyword arguments passed to
            `keras_hub.layers.preprocessing.ImageConverter`.

    Examples:
    ```python
    import keras
    import numpy as np
    from keras_hub.src.layers import ViTImageConverter

    # Example image (replace with your actual image data)
    image = np.random.rand(1, 224, 224, 3)  # Example: (B, H, W, C)

    # Create a ViTImageConverter instance
    converter = ViTImageConverter(
        image_size=(28,28),
        scale=1/255.
    )
    # Preprocess the image
    preprocessed_image = converter(image)
    ```
    """

    backbone_cls = ViTBackbone

    def __init__(
        self, norm_mean=[0.5, 0.5, 0.5], norm_std=[0.5, 0.5, 0.5], **kwargs
    ):
        super().__init__(**kwargs)
        self.norm_mean = norm_mean
        self.norm_std = norm_std

    @preprocessing_function
    def call(self, inputs):
        x = super().call(inputs)
        # By default normalize using imagenet mean and std
        if self.norm_mean:
            x = x - self._expand_non_channel_dims(self.norm_mean, x)
        if self.norm_std:
            x = x / self._expand_non_channel_dims(self.norm_std, x)

        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "norm_mean": self.norm_mean,
                "norm_std": self.norm_std,
            }
        )
        return config



from keras_hub.src.models.mobilenet.mobilenet_backbone import MobileNetBackbone
from keras_hub.src.models.mobilenet.mobilenet_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, MobileNetBackbone)



import math

import keras
from keras import layers
from keras import ops

from keras_hub.src.layers.modeling.position_embedding import PositionEmbedding
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.utils.keras_utils import gelu_approximate
from keras_hub.src.utils.keras_utils import has_flash_attention_support
from keras_hub.src.utils.keras_utils import standardize_data_format


class AdaptiveLayerNormalization(layers.Layer):
    """Adaptive layer normalization.

    Args:
        embedding_dim: int. The size of each embedding vector.
        num_modulations: int. The number of the modulation parameters. The
            available values are `2`, `6` and `9`. Defaults to `2`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.

    References:
    - [FiLM: Visual Reasoning with a General Conditioning Layer](
    https://arxiv.org/abs/1709.07871).
    - [Scalable Diffusion Models with Transformers](
    https://arxiv.org/abs/2212.09748).
    """

    def __init__(self, hidden_dim, num_modulations=2, **kwargs):
        super().__init__(**kwargs)
        hidden_dim = int(hidden_dim)
        num_modulations = int(num_modulations)
        if num_modulations not in (2, 6, 9):
            raise ValueError(
                "`num_modulations` must be `2`, `6` or `9`. "
                f"Received: num_modulations={num_modulations}"
            )
        self.hidden_dim = hidden_dim
        self.num_modulations = num_modulations

        self.silu = layers.Activation("silu", dtype=self.dtype_policy)
        self.dense = layers.Dense(
            num_modulations * hidden_dim, dtype=self.dtype_policy, name="dense"
        )
        self.norm = layers.LayerNormalization(
            epsilon=1e-6,
            center=False,
            scale=False,
            dtype="float32",
            name="norm",
        )

    def build(self, inputs_shape, embeddings_shape):
        self.silu.build(embeddings_shape)
        self.dense.build(embeddings_shape)
        self.norm.build(inputs_shape)

    def call(self, inputs, embeddings, training=None):
        hidden_states = inputs
        emb = self.dense(self.silu(embeddings), training=training)
        if self.num_modulations == 9:
            (
                shift_msa,
                scale_msa,
                gate_msa,
                shift_mlp,
                scale_mlp,
                gate_mlp,
                shift_msa2,
                scale_msa2,
                gate_msa2,
            ) = ops.split(emb, self.num_modulations, axis=1)
        elif self.num_modulations == 6:
            (
                shift_msa,
                scale_msa,
                gate_msa,
                shift_mlp,
                scale_mlp,
                gate_mlp,
            ) = ops.split(emb, self.num_modulations, axis=1)
        else:
            shift_msa, scale_msa = ops.split(emb, self.num_modulations, axis=1)

        scale_msa = ops.expand_dims(scale_msa, axis=1)
        shift_msa = ops.expand_dims(shift_msa, axis=1)
        norm_hidden_states = ops.cast(
            self.norm(hidden_states, training=training), scale_msa.dtype
        )
        hidden_states = ops.add(
            ops.multiply(norm_hidden_states, ops.add(1.0, scale_msa)), shift_msa
        )

        if self.num_modulations == 9:
            scale_msa2 = ops.expand_dims(scale_msa2, axis=1)
            shift_msa2 = ops.expand_dims(shift_msa2, axis=1)
            hidden_states2 = ops.add(
                ops.multiply(norm_hidden_states, ops.add(1.0, scale_msa2)),
                shift_msa2,
            )
            return (
                hidden_states,
                gate_msa,
                shift_mlp,
                scale_mlp,
                gate_mlp,
                hidden_states2,
                gate_msa2,
            )
        elif self.num_modulations == 6:
            return hidden_states, gate_msa, shift_mlp, scale_mlp, gate_mlp
        else:
            return hidden_states

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "num_modulations": self.num_modulations,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape, embeddings_shape):
        if self.num_modulations == 9:
            return (
                inputs_shape,
                embeddings_shape,
                embeddings_shape,
                embeddings_shape,
                embeddings_shape,
                inputs_shape,
                embeddings_shape,
            )
        elif self.num_modulations == 6:
            return (
                inputs_shape,
                embeddings_shape,
                embeddings_shape,
                embeddings_shape,
                embeddings_shape,
            )
        else:
            return inputs_shape


class MLP(layers.Layer):
    """A MLP block with architecture.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        output_dim: int. The number of units in the output layer.
        activation: str of callable. Activation to use in the hidden layers.
            Default to `None`.
    """

    def __init__(self, hidden_dim, output_dim, activation=None, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.output_dim = int(output_dim)
        self.activation = keras.activations.get(activation)

        self.dense1 = layers.Dense(
            hidden_dim,
            activation=self.activation,
            dtype=self.dtype_policy,
            name="dense1",
        )
        self.dense2 = layers.Dense(
            output_dim,
            activation=None,
            dtype=self.dtype_policy,
            name="dense2",
        )

    def build(self, inputs_shape):
        self.dense1.build(inputs_shape)
        inputs_shape = self.dense1.compute_output_shape(inputs_shape)
        self.dense2.build(inputs_shape)

    def call(self, inputs, training=None):
        x = self.dense1(inputs, training=training)
        return self.dense2(x, training=training)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "output_dim": self.output_dim,
                "activation": keras.activations.serialize(self.activation),
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        outputs_shape = list(inputs_shape)
        outputs_shape[-1] = self.output_dim
        return outputs_shape


class PatchEmbedding(layers.Layer):
    """A layer that converts images into patches.

    Args:
        patch_size: int. The size of one side of each patch.
        hidden_dim: int. The number of units in the hidden layers.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(self, patch_size, hidden_dim, data_format=None, **kwargs):
        super().__init__(**kwargs)
        self.patch_size = int(patch_size)
        self.hidden_dim = int(hidden_dim)
        data_format = standardize_data_format(data_format)

        self.patch_embedding = layers.Conv2D(
            hidden_dim,
            kernel_size=patch_size,
            strides=patch_size,
            data_format=data_format,
            dtype=self.dtype_policy,
            name="patch_embedding",
        )

    def build(self, input_shape):
        self.patch_embedding.build(input_shape)

    def call(self, inputs):
        x = self.patch_embedding(inputs)
        x_shape = ops.shape(x)
        x = ops.reshape(x, (x_shape[0], x_shape[1] * x_shape[2], x_shape[3]))
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "hidden_dim": self.hidden_dim,
            }
        )
        return config


class AdjustablePositionEmbedding(PositionEmbedding):
    """A position embedding layer with adjustable height and width.

    The embedding will be cropped to match the input dimensions.

    Args:
        height: int. The maximum height of the embedding.
        width: int. The maximum width of the embedding.
    """

    def __init__(
        self,
        height,
        width,
        initializer="glorot_uniform",
        **kwargs,
    ):
        height = int(height)
        width = int(width)
        sequence_length = height * width
        super().__init__(sequence_length, initializer, **kwargs)
        self.height = height
        self.width = width

    def call(self, inputs, height=None, width=None):
        height = height or self.height
        width = width or self.width
        shape = ops.shape(inputs)
        feature_length = shape[-1]
        top = ops.cast(ops.floor_divide(self.height - height, 2), "int32")
        left = ops.cast(ops.floor_divide(self.width - width, 2), "int32")
        position_embedding = ops.convert_to_tensor(self.position_embeddings)
        position_embedding = ops.reshape(
            position_embedding, (self.height, self.width, feature_length)
        )
        position_embedding = ops.slice(
            position_embedding,
            (top, left, 0),
            (height, width, feature_length),
        )
        position_embedding = ops.reshape(
            position_embedding, (height * width, feature_length)
        )
        position_embedding = ops.expand_dims(position_embedding, axis=0)
        return position_embedding

    def get_config(self):
        config = super().get_config()
        del config["sequence_length"]
        config.update(
            {
                "height": self.height,
                "width": self.width,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        return input_shape


class TimestepEmbedding(layers.Layer):
    """A layer which learns embedding for input timesteps.

    Args:
        embedding_dim: int. The size of the embedding.
        frequency_dim: int. The size of the frequency.
        max_period: int. Controls the maximum frequency of the embeddings.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.

    Reference:
    - [Denoising Diffusion Probabilistic Models](
    https://arxiv.org/abs/2006.11239).
    """

    def __init__(
        self, embedding_dim, frequency_dim=256, max_period=10000, **kwargs
    ):
        super().__init__(**kwargs)
        self.embedding_dim = int(embedding_dim)
        self.frequency_dim = int(frequency_dim)
        self.max_period = float(max_period)
        # Precomputed `freq`.
        half_frequency_dim = frequency_dim // 2
        self.freq = ops.exp(
            ops.divide(
                ops.multiply(
                    -math.log(max_period),
                    ops.arange(0, half_frequency_dim, dtype="float32"),
                ),
                half_frequency_dim,
            )
        )

        self.mlp = MLP(
            embedding_dim,
            embedding_dim,
            "silu",
            dtype=self.dtype_policy,
            name="mlp",
        )

    def build(self, inputs_shape):
        embedding_shape = list(inputs_shape)[1:]
        embedding_shape.append(self.frequency_dim)
        self.mlp.build(embedding_shape)

    def _create_timestep_embedding(self, inputs):
        compute_dtype = keras.backend.result_type(self.compute_dtype, "float32")
        x = ops.cast(inputs, compute_dtype)
        freqs = ops.cast(self.freq, compute_dtype)
        x = ops.multiply(x, ops.expand_dims(freqs, axis=0))
        embedding = ops.concatenate([ops.cos(x), ops.sin(x)], axis=-1)
        if self.frequency_dim % 2 != 0:
            embedding = ops.pad(embedding, [[0, 0], [0, 1]])
        return ops.cast(embedding, self.compute_dtype)

    def call(self, inputs, training=None):
        timestep_embedding = self._create_timestep_embedding(inputs)
        return self.mlp(timestep_embedding, training=training)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "embedding_dim": self.embedding_dim,
                "frequency_dim": self.frequency_dim,
                "max_period": self.max_period,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        output_shape = list(inputs_shape)[1:]
        output_shape.append(self.embedding_dim)
        return output_shape


def get_qk_norm(qk_norm=None, q_norm_name="q_norm", k_norm_name="k_norm"):
    """Helper function to instantiate `LayerNormalization` layers."""
    q_norm = None
    k_norm = None
    if qk_norm is None:
        pass
    elif qk_norm == "rms_norm":
        q_norm = layers.LayerNormalization(
            epsilon=1e-6, rms_scaling=True, dtype="float32", name=q_norm_name
        )
        k_norm = layers.LayerNormalization(
            epsilon=1e-6, rms_scaling=True, dtype="float32", name=k_norm_name
        )
    else:
        raise NotImplementedError(
            "Supported `qk_norm` are `'rms_norm'` and `None`. "
            f"Received: qk_norm={qk_norm}."
        )
    return q_norm, k_norm


class DismantledBlock(layers.Layer):
    """A dismantled block used to compute pre- and post-attention.

    Args:
        num_heads: int. Number of attention heads.
        hidden_dim: int. The number of units in the hidden layers.
        mlp_ratio: float. The expansion ratio of `MLP`.
        use_projection: bool. Whether to use an attention projection layer at
            the end of the block.
        qk_norm: Optional str. Whether to normalize the query and key tensors.
            Available options are `None` and `"rms_norm"`. Defaults to `None`.
        use_dual_attention: bool. Whether to use a dual attention in the
            block. Defaults to `False`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        num_heads,
        hidden_dim,
        mlp_ratio=4.0,
        use_projection=True,
        qk_norm=None,
        use_dual_attention=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.mlp_ratio = mlp_ratio
        self.use_projection = use_projection
        self.qk_norm = qk_norm
        self.use_dual_attention = use_dual_attention

        head_dim = hidden_dim // num_heads
        self.head_dim = head_dim
        mlp_hidden_dim = int(hidden_dim * mlp_ratio)
        self.mlp_hidden_dim = mlp_hidden_dim

        if use_projection:
            self.ada_layer_norm = AdaptiveLayerNormalization(
                hidden_dim,
                num_modulations=9 if use_dual_attention else 6,
                dtype=self.dtype_policy,
                name="ada_layer_norm",
            )
        else:
            self.ada_layer_norm = AdaptiveLayerNormalization(
                hidden_dim, dtype=self.dtype_policy, name="ada_layer_norm"
            )
        self.attention_qkv = layers.Dense(
            hidden_dim * 3, dtype=self.dtype_policy, name="attention_qkv"
        )
        q_norm, k_norm = get_qk_norm(qk_norm)
        if q_norm is not None:
            self.q_norm = q_norm
            self.k_norm = k_norm
        if use_projection:
            self.attention_proj = layers.Dense(
                hidden_dim, dtype=self.dtype_policy, name="attention_proj"
            )
            self.norm2 = layers.LayerNormalization(
                epsilon=1e-6,
                center=False,
                scale=False,
                dtype="float32",
                name="norm2",
            )
            self.mlp = MLP(
                mlp_hidden_dim,
                hidden_dim,
                gelu_approximate,
                dtype=self.dtype_policy,
                name="mlp",
            )

        if use_dual_attention:
            self.attention_qkv2 = layers.Dense(
                hidden_dim * 3, dtype=self.dtype_policy, name="attention_qkv2"
            )
            q_norm2, k_norm2 = get_qk_norm(qk_norm, "q_norm2", "k_norm2")
            if q_norm is not None:
                self.q_norm2 = q_norm2
                self.k_norm2 = k_norm2
            if use_projection:
                self.attention_proj2 = layers.Dense(
                    hidden_dim, dtype=self.dtype_policy, name="attention_proj2"
                )

    def build(self, inputs_shape, timestep_embedding):
        self.ada_layer_norm.build(inputs_shape, timestep_embedding)
        self.attention_qkv.build(inputs_shape)
        if self.qk_norm is not None:
            # [batch_size, sequence_length, num_heads, head_dim]
            self.q_norm.build([None, None, self.num_heads, self.head_dim])
            self.k_norm.build([None, None, self.num_heads, self.head_dim])
        if self.use_projection:
            self.attention_proj.build(inputs_shape)
            self.norm2.build(inputs_shape)
            self.mlp.build(inputs_shape)
        if self.use_dual_attention:
            self.attention_qkv2.build(inputs_shape)
            if self.qk_norm is not None:
                self.q_norm2.build([None, None, self.num_heads, self.head_dim])
                self.k_norm2.build([None, None, self.num_heads, self.head_dim])
            if self.use_projection:
                self.attention_proj2.build(inputs_shape)

    def _modulate(self, inputs, shift, scale):
        inputs = ops.cast(inputs, self.compute_dtype)
        shift = ops.cast(shift, self.compute_dtype)
        scale = ops.cast(scale, self.compute_dtype)
        return ops.add(ops.multiply(inputs, ops.add(scale, 1.0)), shift)

    def _compute_pre_attention(self, inputs, timestep_embedding, training=None):
        batch_size = ops.shape(inputs)[0]
        if self.use_projection:
            x, gate_msa, shift_mlp, scale_mlp, gate_mlp = self.ada_layer_norm(
                inputs, timestep_embedding, training=training
            )
            qkv = self.attention_qkv(x, training=training)
            qkv = ops.reshape(
                qkv, (batch_size, -1, 3, self.num_heads, self.head_dim)
            )
            q, k, v = ops.unstack(qkv, 3, axis=2)
            if self.qk_norm is not None:
                q = ops.cast(
                    self.q_norm(q, training=training), self.compute_dtype
                )
                k = ops.cast(
                    self.k_norm(k, training=training), self.compute_dtype
                )
            return (q, k, v), (inputs, gate_msa, shift_mlp, scale_mlp, gate_mlp)
        else:
            x = self.ada_layer_norm(
                inputs, timestep_embedding, training=training
            )
            qkv = self.attention_qkv(x, training=training)
            qkv = ops.reshape(
                qkv, (batch_size, -1, 3, self.num_heads, self.head_dim)
            )
            q, k, v = ops.unstack(qkv, 3, axis=2)
            if self.qk_norm is not None:
                q = ops.cast(
                    self.q_norm(q, training=training), self.compute_dtype
                )
                k = ops.cast(
                    self.k_norm(k, training=training), self.compute_dtype
                )
            return (q, k, v)

    def _compute_post_attention(
        self, inputs, inputs_intermediates, training=None
    ):
        x, gate_msa, shift_mlp, scale_mlp, gate_mlp = inputs_intermediates
        gate_msa = ops.expand_dims(gate_msa, axis=1)
        shift_mlp = ops.expand_dims(shift_mlp, axis=1)
        scale_mlp = ops.expand_dims(scale_mlp, axis=1)
        gate_mlp = ops.expand_dims(gate_mlp, axis=1)
        attn = self.attention_proj(inputs, training=training)
        x = ops.add(x, ops.multiply(gate_msa, attn))
        x = ops.add(
            x,
            ops.multiply(
                gate_mlp,
                self.mlp(
                    self._modulate(self.norm2(x), shift_mlp, scale_mlp),
                    training=training,
                ),
            ),
        )
        return x

    def _compute_pre_attention_with_dual_attention(
        self, inputs, timestep_embedding, training=None
    ):
        batch_size = ops.shape(inputs)[0]
        x, gate_msa, shift_mlp, scale_mlp, gate_mlp, x2, gate_msa2 = (
            self.ada_layer_norm(inputs, timestep_embedding, training=training)
        )
        # Compute the main attention
        qkv = self.attention_qkv(x, training=training)
        qkv = ops.reshape(
            qkv, (batch_size, -1, 3, self.num_heads, self.head_dim)
        )
        q, k, v = ops.unstack(qkv, 3, axis=2)
        if self.qk_norm is not None:
            q = ops.cast(self.q_norm(q, training=training), self.compute_dtype)
            k = ops.cast(self.k_norm(k, training=training), self.compute_dtype)
        # Compute the dual attention
        qkv2 = self.attention_qkv2(x2, training=training)
        qkv2 = ops.reshape(
            qkv2, (batch_size, -1, 3, self.num_heads, self.head_dim)
        )
        q2, k2, v2 = ops.unstack(qkv2, 3, axis=2)
        if self.qk_norm is not None:
            q2 = ops.cast(
                self.q_norm2(q2, training=training), self.compute_dtype
            )
            k2 = ops.cast(
                self.k_norm2(k2, training=training), self.compute_dtype
            )
        return (
            (q, k, v),
            (q2, k2, v2),
            (inputs, gate_msa, shift_mlp, scale_mlp, gate_mlp, gate_msa2),
        )

    def _compute_post_attention_with_dual_attention(
        self, inputs, inputs2, inputs_intermediates, training=None
    ):
        x, gate_msa, shift_mlp, scale_mlp, gate_mlp, gate_msa2 = (
            inputs_intermediates
        )
        gate_msa = ops.expand_dims(gate_msa, axis=1)
        shift_mlp = ops.expand_dims(shift_mlp, axis=1)
        scale_mlp = ops.expand_dims(scale_mlp, axis=1)
        gate_mlp = ops.expand_dims(gate_mlp, axis=1)
        gate_msa2 = ops.expand_dims(gate_msa2, axis=1)
        attn = self.attention_proj(inputs, training=training)
        x = ops.add(x, ops.multiply(gate_msa, attn))
        attn2 = self.attention_proj2(inputs2, training=training)
        x = ops.add(x, ops.multiply(gate_msa2, attn2))
        x = ops.add(
            x,
            ops.multiply(
                gate_mlp,
                self.mlp(
                    self._modulate(self.norm2(x), shift_mlp, scale_mlp),
                    training=training,
                ),
            ),
        )
        return x

    def call(
        self,
        inputs,
        timestep_embedding=None,
        inputs_intermediates=None,
        inputs2=None,  # For the dual attention.
        pre_attention=True,
        training=None,
    ):
        if pre_attention:
            if self.use_dual_attention:
                return self._compute_pre_attention_with_dual_attention(
                    inputs, timestep_embedding, training=training
                )
            else:
                return self._compute_pre_attention(
                    inputs, timestep_embedding, training=training
                )
        else:
            if self.use_dual_attention:
                return self._compute_post_attention_with_dual_attention(
                    inputs, inputs2, inputs_intermediates, training=training
                )
            else:
                return self._compute_post_attention(
                    inputs, inputs_intermediates, training=training
                )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_heads": self.num_heads,
                "hidden_dim": self.hidden_dim,
                "mlp_ratio": self.mlp_ratio,
                "use_projection": self.use_projection,
                "qk_norm": self.qk_norm,
                "use_dual_attention": self.use_dual_attention,
            }
        )
        return config


class MMDiTBlock(layers.Layer):
    """A MMDiT block consisting of two `DismantledBlock` layers.

    One `DismantledBlock` processes the input latents, and the other processes
    the context embedding. This block integrates two modalities within the
    attention operation, allowing each representation to operate in its own
    space while considering the other.

    Args:
        num_heads: int. Number of attention heads.
        hidden_dim: int. The number of units in the hidden layers.
        mlp_ratio: float. The expansion ratio of `MLP`.
        use_context_projection: bool. Whether to use an attention projection
            layer at the end of the context block.
        qk_norm: Optional str. Whether to normalize the query and key tensors.
            Available options are `None` and `"rms_norm"`. Defaults to `None`.
        use_dual_attention: bool. Whether to use a dual attention in the
            block. Defaults to `False`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.

    Reference:
    - [Scaling Rectified Flow Transformers for High-Resolution Image Synthesis](
    https://arxiv.org/abs/2403.03206)
    """

    def __init__(
        self,
        num_heads,
        hidden_dim,
        mlp_ratio=4.0,
        use_context_projection=True,
        qk_norm=None,
        use_dual_attention=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.mlp_ratio = mlp_ratio
        self.use_context_projection = use_context_projection
        self.qk_norm = qk_norm
        self.use_dual_attention = use_dual_attention

        head_dim = hidden_dim // num_heads
        self.head_dim = head_dim
        self._inverse_sqrt_key_dim = 1.0 / math.sqrt(head_dim)

        self.x_block = DismantledBlock(
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_ratio=mlp_ratio,
            use_projection=True,
            qk_norm=qk_norm,
            use_dual_attention=use_dual_attention,
            dtype=self.dtype_policy,
            name="x_block",
        )
        self.context_block = DismantledBlock(
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_ratio=mlp_ratio,
            use_projection=use_context_projection,
            qk_norm=qk_norm,
            dtype=self.dtype_policy,
            name="context_block",
        )
        self.softmax = layers.Softmax(dtype="float32")

    def build(self, inputs_shape, context_shape, timestep_embedding_shape):
        self.x_block.build(inputs_shape, timestep_embedding_shape)
        self.context_block.build(context_shape, timestep_embedding_shape)

    def _compute_attention(self, query, key, value):
        batch_size = ops.shape(query)[0]

        if has_flash_attention_support():
            # Use `dot_product_attention` with Flash Attention support if
            # available.
            encoded = ops.dot_product_attention(
                query,
                key,
                value,
                scale=self._inverse_sqrt_key_dim,
            )
            return ops.reshape(
                encoded, (batch_size, -1, self.num_heads * self.head_dim)
            )

        # Ref: jax.nn.dot_product_attention
        # https://github.com/jax-ml/jax/blob/db89c245ac66911c98f265a05956fdfa4bc79d83/jax/_src/nn/functions.py#L846
        logits = ops.einsum("BTNH,BSNH->BNTS", query, key)
        logits = ops.multiply(logits, self._inverse_sqrt_key_dim)
        probs = self.softmax(logits)
        probs = ops.cast(probs, self.compute_dtype)
        encoded = ops.einsum("BNTS,BSNH->BTNH", probs, value)
        return ops.reshape(
            encoded, (batch_size, -1, self.num_heads * self.head_dim)
        )

    def call(self, inputs, context, timestep_embedding, training=None):
        # Compute pre-attention.
        x = inputs
        if self.use_context_projection:
            context_qkv, context_intermediates = self.context_block(
                context,
                timestep_embedding=timestep_embedding,
                training=training,
            )
        else:
            context_qkv = self.context_block(
                context,
                timestep_embedding=timestep_embedding,
                training=training,
            )
        context_len = ops.shape(context_qkv[0])[1]
        if self.x_block.use_dual_attention:
            x_qkv, x_qkv2, x_intermediates = self.x_block(
                x, timestep_embedding=timestep_embedding, training=training
            )
        else:
            x_qkv, x_intermediates = self.x_block(
                x, timestep_embedding=timestep_embedding, training=training
            )
        q = ops.concatenate([context_qkv[0], x_qkv[0]], axis=1)
        k = ops.concatenate([context_qkv[1], x_qkv[1]], axis=1)
        v = ops.concatenate([context_qkv[2], x_qkv[2]], axis=1)

        # Compute attention.
        attention = self._compute_attention(q, k, v)
        context_attention = attention[:, :context_len]
        x_attention = attention[:, context_len:]

        # Compute post-attention.
        if self.x_block.use_dual_attention:
            q2, k2, v2 = x_qkv2
            x_attention2 = self._compute_attention(q2, k2, v2)
            x = self.x_block(
                x_attention,
                inputs_intermediates=x_intermediates,
                inputs2=x_attention2,
                pre_attention=False,
                training=training,
            )
        else:
            x = self.x_block(
                x_attention,
                inputs_intermediates=x_intermediates,
                pre_attention=False,
                training=training,
            )
        if self.use_context_projection:
            context = self.context_block(
                context_attention,
                inputs_intermediates=context_intermediates,
                pre_attention=False,
                training=training,
            )
            return x, context
        else:
            return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_heads": self.num_heads,
                "hidden_dim": self.hidden_dim,
                "mlp_ratio": self.mlp_ratio,
                "use_context_projection": self.use_context_projection,
                "qk_norm": self.qk_norm,
                "use_dual_attention": self.use_dual_attention,
            }
        )
        return config

    def compute_output_shape(
        self, inputs_shape, context_shape, timestep_embedding_shape
    ):
        if self.use_context_projection:
            return inputs_shape, context_shape
        else:
            return inputs_shape


class Unpatch(layers.Layer):
    """A layer that reconstructs the image from hidden patches.

    Args:
        patch_size: int. The size of each square patch in the input image.
        output_dim: int. The number of units in the output layer.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(self, patch_size, output_dim, **kwargs):
        super().__init__(**kwargs)
        self.patch_size = int(patch_size)
        self.output_dim = int(output_dim)

    def call(self, inputs, height, width):
        patch_size = self.patch_size
        output_dim = self.output_dim
        x = ops.reshape(
            inputs,
            (-1, height, width, patch_size, patch_size, output_dim),
        )
        # (b, h, w, p1, p2, o) -> (b, h, p1, w, p2, o)
        x = ops.transpose(x, (0, 1, 3, 2, 4, 5))
        return ops.reshape(
            x,
            (-1, height * patch_size, width * patch_size, output_dim),
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "output_dim": self.output_dim,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        inputs_shape = list(inputs_shape)
        return [inputs_shape[0], None, None, self.output_dim]


class MMDiT(Backbone):
    """A Multimodal Diffusion Transformer (MMDiT) model.

    MMDiT is introduced in [
    Scaling Rectified Flow Transformers for High-Resolution Image Synthesis](
    https://arxiv.org/abs/2403.03206).

    Args:
        patch_size: int. The size of each square patch in the input image.
        hidden_dim: int. The size of the transformer hidden state at the end
            of each transformer layer.
        num_layers: int. The number of transformer layers.
        num_heads: int. The number of attention heads for each transformer.
        position_size: int. The size of the height and width for the position
            embedding.
        mlp_ratio: float. The ratio of the mlp hidden dim to the transformer
        latent_shape: tuple. The shape of the latent image.
        context_shape: tuple. The shape of the context.
        pooled_projection_shape: tuple. The shape of the pooled projection.
        qk_norm: Optional str. Whether to normalize the query and key tensors in
            the intermediate blocks. Available options are `None` and
            `"rms_norm"`. Defaults to `None`.
        dual_attention_indices: Optional tuple. Specifies the indices of
            the blocks that serve as dual attention blocks. Typically, this is
            for 3.5 version. Defaults to `None`.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.
    """

    def __init__(
        self,
        patch_size,
        hidden_dim,
        num_layers,
        num_heads,
        position_size,
        mlp_ratio=4.0,
        latent_shape=(64, 64, 16),
        context_shape=(None, 4096),
        pooled_projection_shape=(2048,),
        qk_norm=None,
        dual_attention_indices=None,
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        if None in latent_shape:
            raise ValueError(
                "`latent_shape` must be fully specified. "
                f"Received: latent_shape={latent_shape}"
            )
        image_height = latent_shape[0] // patch_size
        image_width = latent_shape[1] // patch_size
        output_dim = latent_shape[-1]
        output_dim_in_final = patch_size**2 * output_dim
        dual_attention_indices = dual_attention_indices or ()
        data_format = standardize_data_format(data_format)
        if data_format != "channels_last":
            raise NotImplementedError(
                "Currently only 'channels_last' is supported."
            )

        # === Layers ===
        self.patch_embedding = PatchEmbedding(
            patch_size,
            hidden_dim,
            data_format=data_format,
            dtype=dtype,
            name="patch_embedding",
        )
        self.position_embedding_add = layers.Add(
            dtype=dtype, name="position_embedding_add"
        )
        self.position_embedding = AdjustablePositionEmbedding(
            position_size, position_size, dtype=dtype, name="position_embedding"
        )
        self.context_embedding = layers.Dense(
            hidden_dim,
            dtype=dtype,
            name="context_embedding",
        )
        self.vector_embedding = MLP(
            hidden_dim, hidden_dim, "silu", dtype=dtype, name="vector_embedding"
        )
        self.vector_embedding_add = layers.Add(
            dtype=dtype, name="vector_embedding_add"
        )
        self.timestep_embedding = TimestepEmbedding(
            hidden_dim, dtype=dtype, name="timestep_embedding"
        )
        self.joint_blocks = [
            MMDiTBlock(
                num_heads,
                hidden_dim,
                mlp_ratio,
                use_context_projection=not (i == num_layers - 1),
                qk_norm=qk_norm,
                use_dual_attention=i in dual_attention_indices,
                dtype=dtype,
                name=f"joint_block_{i}",
            )
            for i in range(num_layers)
        ]
        self.output_ada_layer_norm = AdaptiveLayerNormalization(
            hidden_dim, dtype=dtype, name="output_ada_layer_norm"
        )
        self.output_dense = layers.Dense(
            output_dim_in_final, dtype=dtype, name="output_dense"
        )
        self.unpatch = Unpatch(
            patch_size, output_dim, dtype=dtype, name="unpatch"
        )

        # === Functional Model ===
        latent_inputs = layers.Input(shape=latent_shape, name="latent")
        context_inputs = layers.Input(shape=context_shape, name="context")
        pooled_projection_inputs = layers.Input(
            shape=pooled_projection_shape, name="pooled_projection"
        )
        timestep_inputs = layers.Input(shape=(1,), name="timestep")

        # Embeddings.
        x = self.patch_embedding(latent_inputs)
        position_embedding = self.position_embedding(
            x, height=image_height, width=image_width
        )
        x = self.position_embedding_add([x, position_embedding])
        context = self.context_embedding(context_inputs)
        pooled_projection = self.vector_embedding(pooled_projection_inputs)
        timestep_embedding = self.timestep_embedding(timestep_inputs)
        timestep_embedding = self.vector_embedding_add(
            [timestep_embedding, pooled_projection]
        )

        # Blocks.
        for block in self.joint_blocks:
            if block.use_context_projection:
                x, context = block(x, context, timestep_embedding)
            else:
                x = block(x, context, timestep_embedding)

        # Output layer.
        x = self.output_ada_layer_norm(x, timestep_embedding)
        x = self.output_dense(x)
        outputs = self.unpatch(x, height=image_height, width=image_width)

        super().__init__(
            inputs={
                "latent": latent_inputs,
                "context": context_inputs,
                "pooled_projection": pooled_projection_inputs,
                "timestep": timestep_inputs,
            },
            outputs=outputs,
            **kwargs,
        )

        # === Config ===
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.position_size = position_size
        self.mlp_ratio = mlp_ratio
        self.latent_shape = latent_shape
        self.context_shape = context_shape
        self.pooled_projection_shape = pooled_projection_shape
        self.qk_norm = qk_norm
        self.dual_attention_indices = dual_attention_indices

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "patch_size": self.patch_size,
                "hidden_dim": self.hidden_dim,
                "num_layers": self.num_layers,
                "num_heads": self.num_heads,
                "position_size": self.position_size,
                "mlp_ratio": self.mlp_ratio,
                "latent_shape": self.latent_shape,
                "context_shape": self.context_shape,
                "pooled_projection_shape": self.pooled_projection_shape,
                "qk_norm": self.qk_norm,
                "dual_attention_indices": self.dual_attention_indices,
            }
        )
        return config



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.resnet.resnet_backbone import (
    apply_basic_block as resnet_basic_block,
)


@keras_hub_export("keras_hub.models.BASNetBackbone")
class BASNetBackbone(Backbone):
    """BASNet architecture for semantic segmentation.

    A Keras model implementing the BASNet architecture described in [BASNet:
    Boundary-Aware Segmentation Network for Mobile and Web Applications](
    https://arxiv.org/abs/2101.04704). BASNet uses a predict-refine
    architecture for highly accurate image segmentation.

    Args:
        image_encoder: A `keras_hub.models.ResNetBackbone` instance. The
            backbone network for the model that is used as a feature extractor
            for BASNet prediction encoder. Currently supported backbones are
            ResNet18 and ResNet34.
            (Note: Do not specify `image_shape` within the backbone.
            Please provide these while initializing the 'BASNetBackbone' model)
        num_classes: int, the number of classes for the segmentation model.
        image_shape: optional shape tuple, defaults to (None, None, 3).
        projection_filters: int, number of filters in the convolution layer
            projecting low-level features from the `backbone`.
        prediction_heads: (Optional) List of `keras.layers.Layer` defining
            the prediction module head for the model. If not provided, a
            default head is created with a Conv2D layer followed by resizing.
        refinement_head: (Optional) a `keras.layers.Layer` defining the
            refinement module head for the model. If not provided, a default
            head is created with a Conv2D layer.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.
    """

    def __init__(
        self,
        image_encoder,
        num_classes,
        image_shape=(None, None, 3),
        projection_filters=64,
        prediction_heads=None,
        refinement_head=None,
        dtype=None,
        **kwargs,
    ):
        if not isinstance(image_encoder, keras.layers.Layer) or not isinstance(
            image_encoder, keras.Model
        ):
            raise ValueError(
                "Argument `image_encoder` must be a `keras.layers.Layer`"
                f" instance or `keras.Model`. Received instead"
                f" image_encoder={image_encoder} (of type"
                f" {type(image_encoder)})."
            )

        if tuple(image_encoder.image_shape) != (None, None, 3):
            raise ValueError(
                "Do not specify `image_shape` within the"
                " `BASNetBackbone`'s image_encoder. \nPlease provide"
                " `image_shape` while initializing the 'BASNetBackbone' model."
            )

        # === Functional Model ===
        inputs = keras.layers.Input(shape=image_shape)
        x = inputs

        if prediction_heads is None:
            prediction_heads = []
            for size in (1, 2, 4, 8, 16, 32, 32):
                head_layers = [
                    keras.layers.Conv2D(
                        num_classes,
                        kernel_size=(3, 3),
                        padding="same",
                        dtype=dtype,
                    )
                ]
                if size != 1:
                    head_layers.append(
                        keras.layers.UpSampling2D(
                            size=size, interpolation="bilinear", dtype=dtype
                        )
                    )
                prediction_heads.append(keras.Sequential(head_layers))

        if refinement_head is None:
            refinement_head = keras.Sequential(
                [
                    keras.layers.Conv2D(
                        num_classes,
                        kernel_size=(3, 3),
                        padding="same",
                        dtype=dtype,
                    ),
                ]
            )

        # Prediction model.
        predict_model = basnet_predict(
            x, image_encoder, projection_filters, prediction_heads, dtype=dtype
        )

        # Refinement model.
        refine_model = basnet_rrm(
            predict_model, projection_filters, refinement_head, dtype=dtype
        )

        outputs = refine_model.outputs  # Combine outputs.
        outputs.extend(predict_model.outputs)

        output_names = ["refine_out"] + [
            f"predict_out_{i}" for i in range(1, len(outputs))
        ]

        outputs = {
            output_name: keras.layers.Activation(
                "sigmoid", name=output_name, dtype=dtype
            )(output)
            for output, output_name in zip(outputs, output_names)
        }

        super().__init__(inputs=inputs, outputs=outputs, dtype=dtype, **kwargs)

        # === Config ===
        self.image_encoder = image_encoder
        self.num_classes = num_classes
        self.image_shape = image_shape
        self.projection_filters = projection_filters
        self.prediction_heads = prediction_heads
        self.refinement_head = refinement_head

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "image_encoder": keras.saving.serialize_keras_object(
                    self.image_encoder
                ),
                "num_classes": self.num_classes,
                "image_shape": self.image_shape,
                "projection_filters": self.projection_filters,
                "prediction_heads": [
                    keras.saving.serialize_keras_object(prediction_head)
                    for prediction_head in self.prediction_heads
                ],
                "refinement_head": keras.saving.serialize_keras_object(
                    self.refinement_head
                ),
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        if "image_encoder" in config:
            config["image_encoder"] = keras.layers.deserialize(
                config["image_encoder"]
            )
        if "prediction_heads" in config and isinstance(
            config["prediction_heads"], list
        ):
            for i in range(len(config["prediction_heads"])):
                if isinstance(config["prediction_heads"][i], dict):
                    config["prediction_heads"][i] = keras.layers.deserialize(
                        config["prediction_heads"][i]
                    )

        if "refinement_head" in config and isinstance(
            config["refinement_head"], dict
        ):
            config["refinement_head"] = keras.layers.deserialize(
                config["refinement_head"]
            )
        return super().from_config(config)


def convolution_block(x_input, filters, dilation=1, dtype=None):
    """Apply convolution + batch normalization + ReLU activation.

    Args:
        x_input: Input keras tensor.
        filters: int, number of output filters in the convolution.
        dilation: int, dilation rate for the convolution operation.
            Defaults to 1.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.

    Returns:
        A tensor with convolution, batch normalization, and ReLU
        activation applied.
    """
    x = keras.layers.Conv2D(
        filters, (3, 3), padding="same", dilation_rate=dilation, dtype=dtype
    )(x_input)
    x = keras.layers.BatchNormalization(dtype=dtype)(x)
    return keras.layers.Activation("relu", dtype=dtype)(x)


def get_resnet_block(_resnet, block_num):
    """Extract and return a specific ResNet block.

    Args:
        _resnet: `keras.Model`. ResNet model instance.
        block_num: int, block number to extract.

    Returns:
        A Keras Model representing the specified ResNet block.
    """

    extractor_levels = ["P2", "P3", "P4", "P5"]
    num_blocks = _resnet.stackwise_num_blocks
    if block_num == 0:
        x = _resnet.get_layer("pool1_pool").output
    else:
        x = _resnet.pyramid_outputs[extractor_levels[block_num - 1]]
    y = _resnet.get_layer(
        f"stack{block_num}_block{num_blocks[block_num] - 1}_add"
    ).output
    return keras.models.Model(
        inputs=x,
        outputs=y,
        name=f"resnet_block{block_num + 1}",
    )


def basnet_predict(x_input, backbone, filters, segmentation_heads, dtype=None):
    """BASNet Prediction Module.

    This module outputs a coarse label map by integrating heavy
    encoder, bridge, and decoder blocks.

    Args:
        x_input: Input keras tensor.
        backbone: `keras.Model`. The backbone network used as a feature
            extractor for BASNet prediction encoder.
        filters: int, the number of filters.
        segmentation_heads: List of `keras.layers.Layer`, A list of Keras
            layers serving as the segmentation head for prediction module.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.


    Returns:
        A Keras Model that integrates the encoder, bridge, and decoder
        blocks for coarse label map prediction.
    """
    num_stages = 6

    x = x_input

    # -------------Encoder--------------
    x = keras.layers.Conv2D(
        filters, kernel_size=(3, 3), padding="same", dtype=dtype
    )(x)

    encoder_blocks = []
    for i in range(num_stages):
        if i < 4:  # First four stages are adopted from ResNet backbone.
            x = get_resnet_block(backbone, i)(x)
            encoder_blocks.append(x)
        else:  # Last 2 stages consist of three basic resnet blocks.
            x = keras.layers.MaxPool2D(
                pool_size=(2, 2), strides=(2, 2), dtype=dtype
            )(x)
            for j in range(3):
                x = resnet_basic_block(
                    x,
                    filters=x.shape[3],
                    conv_shortcut=False,
                    name=f"v1_basic_block_{i + 1}_{j + 1}",
                    dtype=dtype,
                )
            encoder_blocks.append(x)

    # -------------Bridge-------------
    x = convolution_block(x, filters=filters * 8, dilation=2, dtype=dtype)
    x = convolution_block(x, filters=filters * 8, dilation=2, dtype=dtype)
    x = convolution_block(x, filters=filters * 8, dilation=2, dtype=dtype)
    encoder_blocks.append(x)

    # -------------Decoder-------------
    decoder_blocks = []
    for i in reversed(range(num_stages)):
        if i != (num_stages - 1):  # Except first, scale other decoder stages.
            x = keras.layers.UpSampling2D(
                size=2, interpolation="bilinear", dtype=dtype
            )(x)

        x = keras.layers.concatenate([encoder_blocks[i], x], axis=-1)
        x = convolution_block(x, filters=filters * 8, dtype=dtype)
        x = convolution_block(x, filters=filters * 8, dtype=dtype)
        x = convolution_block(x, filters=filters * 8, dtype=dtype)
        decoder_blocks.append(x)

    decoder_blocks.reverse()  # Change order from last to first decoder stage.
    decoder_blocks.append(encoder_blocks[-1])  # Copy bridge to decoder.

    # -------------Side Outputs--------------
    decoder_blocks = [
        segmentation_head(decoder_block)  # Prediction segmentation head.
        for segmentation_head, decoder_block in zip(
            segmentation_heads, decoder_blocks
        )
    ]

    return keras.models.Model(inputs=[x_input], outputs=decoder_blocks)


def basnet_rrm(base_model, filters, segmentation_head, dtype=None):
    """BASNet Residual Refinement Module (RRM).

    This module outputs a fine label map by integrating light encoder,
    bridge, and decoder blocks.

    Args:
        base_model: Keras model used as the base or coarse label map.
        filters: int, the number of filters.
        segmentation_head: a `keras.layers.Layer`, A Keras layer serving
            as the segmentation head for refinement module.
        dtype: `None` or str or `keras.mixed_precision.DTypePolicy`. The dtype
            to use for the model's computations and weights.

    Returns:
        A Keras Model that constructs the Residual Refinement Module (RRM).
    """
    num_stages = 4

    x_input = base_model.output[0]

    # -------------Encoder--------------
    x = keras.layers.Conv2D(
        filters, kernel_size=(3, 3), padding="same", dtype=dtype
    )(x_input)

    encoder_blocks = []
    for _ in range(num_stages):
        x = convolution_block(x, filters=filters)
        encoder_blocks.append(x)
        x = keras.layers.MaxPool2D(
            pool_size=(2, 2), strides=(2, 2), dtype=dtype
        )(x)

    # -------------Bridge--------------
    x = convolution_block(x, filters=filters, dtype=dtype)

    # -------------Decoder--------------
    for i in reversed(range(num_stages)):
        x = keras.layers.UpSampling2D(
            size=2, interpolation="bilinear", dtype=dtype
        )(x)
        x = keras.layers.concatenate([encoder_blocks[i], x], axis=-1)
        x = convolution_block(x, filters=filters)

    x = segmentation_head(x)  # Refinement segmentation head.

    # ------------- refined = coarse + residual
    x = keras.layers.Add(dtype=dtype)(
        [x_input, x]
    )  # Add prediction + refinement output

    return keras.models.Model(inputs=base_model.input, outputs=[x])



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_hub.src.models.albert.albert_backbone import AlbertBackbone
from keras_hub.src.models.albert.albert_masked_lm import AlbertMaskedLM
from keras_hub.src.models.albert.albert_masked_lm_preprocessor import (
    AlbertMaskedLMPreprocessor,
)
from keras_hub.src.models.albert.albert_text_classifier import (
    AlbertTextClassifier,
)
from keras_hub.src.models.albert.albert_text_classifier import (
    AlbertTextClassifier as AlbertClassifier,
)
from keras_hub.src.models.albert.albert_text_classifier_preprocessor import (
    AlbertTextClassifierPreprocessor,
)
from keras_hub.src.models.albert.albert_text_classifier_preprocessor import (
    AlbertTextClassifierPreprocessor as AlbertPreprocessor,
)
from keras_hub.src.models.albert.albert_tokenizer import AlbertTokenizer
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.bart.bart_backbone import BartBackbone
from keras_hub.src.models.bart.bart_seq_2_seq_lm import BartSeq2SeqLM
from keras_hub.src.models.bart.bart_seq_2_seq_lm_preprocessor import (
    BartSeq2SeqLMPreprocessor,
)
from keras_hub.src.models.bart.bart_tokenizer import BartTokenizer
from keras_hub.src.models.basnet.basnet import BASNetImageSegmenter
from keras_hub.src.models.basnet.basnet_backbone import BASNetBackbone
from keras_hub.src.models.basnet.basnet_preprocessor import BASNetPreprocessor
from keras_hub.src.models.bert.bert_backbone import BertBackbone
from keras_hub.src.models.bert.bert_masked_lm import BertMaskedLM
from keras_hub.src.models.bert.bert_masked_lm_preprocessor import (
    BertMaskedLMPreprocessor,
)
from keras_hub.src.models.bert.bert_text_classifier import BertTextClassifier
from keras_hub.src.models.bert.bert_text_classifier import (
    BertTextClassifier as BertClassifier,
)
from keras_hub.src.models.bert.bert_text_classifier_preprocessor import (
    BertTextClassifierPreprocessor,
)
from keras_hub.src.models.bert.bert_text_classifier_preprocessor import (
    BertTextClassifierPreprocessor as BertPreprocessor,
)
from keras_hub.src.models.bert.bert_tokenizer import BertTokenizer
from keras_hub.src.models.bloom.bloom_backbone import BloomBackbone
from keras_hub.src.models.bloom.bloom_causal_lm import BloomCausalLM
from keras_hub.src.models.bloom.bloom_causal_lm_preprocessor import (
    BloomCausalLMPreprocessor,
)
from keras_hub.src.models.bloom.bloom_tokenizer import BloomTokenizer
from keras_hub.src.models.causal_lm import CausalLM
from keras_hub.src.models.causal_lm_preprocessor import CausalLMPreprocessor
from keras_hub.src.models.clip.clip_backbone import CLIPBackbone
from keras_hub.src.models.clip.clip_preprocessor import CLIPPreprocessor
from keras_hub.src.models.clip.clip_text_encoder import CLIPTextEncoder
from keras_hub.src.models.clip.clip_tokenizer import CLIPTokenizer
from keras_hub.src.models.clip.clip_vision_encoder import CLIPVisionEncoder
from keras_hub.src.models.csp_darknet.csp_darknet_backbone import (
    CSPDarkNetBackbone,
)
from keras_hub.src.models.csp_darknet.csp_darknet_image_classifier import (
    CSPDarkNetImageClassifier,
)
from keras_hub.src.models.deberta_v3.deberta_v3_backbone import (
    DebertaV3Backbone,
)
from keras_hub.src.models.deberta_v3.deberta_v3_masked_lm import (
    DebertaV3MaskedLM,
)
from keras_hub.src.models.deberta_v3.deberta_v3_masked_lm_preprocessor import (
    DebertaV3MaskedLMPreprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier import (
    DebertaV3TextClassifier,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier import (
    DebertaV3TextClassifier as DebertaV3Classifier,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier_preprocessor import (
    DebertaV3TextClassifierPreprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_text_classifier_preprocessor import (
    DebertaV3TextClassifierPreprocessor as DebertaV3Preprocessor,
)
from keras_hub.src.models.deberta_v3.deberta_v3_tokenizer import (
    DebertaV3Tokenizer,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_backbone import (
    DeepLabV3Backbone,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_image_segmeter_preprocessor import (
    DeepLabV3ImageSegmenterPreprocessor,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_segmenter import (
    DeepLabV3ImageSegmenter,
)
from keras_hub.src.models.densenet.densenet_backbone import DenseNetBackbone
from keras_hub.src.models.densenet.densenet_image_classifier import (
    DenseNetImageClassifier,
)
from keras_hub.src.models.densenet.densenet_image_classifier_preprocessor import (
    DenseNetImageClassifierPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_backbone import (
    DistilBertBackbone,
)
from keras_hub.src.models.distil_bert.distil_bert_masked_lm import (
    DistilBertMaskedLM,
)
from keras_hub.src.models.distil_bert.distil_bert_masked_lm_preprocessor import (
    DistilBertMaskedLMPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier import (
    DistilBertTextClassifier,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier import (
    DistilBertTextClassifier as DistilBertClassifier,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier_preprocessor import (
    DistilBertTextClassifierPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_text_classifier_preprocessor import (
    DistilBertTextClassifierPreprocessor as DistilBertPreprocessor,
)
from keras_hub.src.models.distil_bert.distil_bert_tokenizer import (
    DistilBertTokenizer,
)
from keras_hub.src.models.efficientnet.efficientnet_backbone import (
    EfficientNetBackbone,
)
from keras_hub.src.models.efficientnet.efficientnet_image_classifier import (
    EfficientNetImageClassifier,
)
from keras_hub.src.models.efficientnet.efficientnet_image_classifier_preprocessor import (
    EfficientNetImageClassifierPreprocessor,
)
from keras_hub.src.models.electra.electra_backbone import ElectraBackbone
from keras_hub.src.models.electra.electra_tokenizer import ElectraTokenizer
from keras_hub.src.models.f_net.f_net_backbone import FNetBackbone
from keras_hub.src.models.f_net.f_net_masked_lm import FNetMaskedLM
from keras_hub.src.models.f_net.f_net_masked_lm_preprocessor import (
    FNetMaskedLMPreprocessor,
)
from keras_hub.src.models.f_net.f_net_text_classifier import FNetTextClassifier
from keras_hub.src.models.f_net.f_net_text_classifier import (
    FNetTextClassifier as FNetClassifier,
)
from keras_hub.src.models.f_net.f_net_text_classifier_preprocessor import (
    FNetTextClassifierPreprocessor,
)
from keras_hub.src.models.f_net.f_net_text_classifier_preprocessor import (
    FNetTextClassifierPreprocessor as FNetPreprocessor,
)
from keras_hub.src.models.f_net.f_net_tokenizer import FNetTokenizer
from keras_hub.src.models.falcon.falcon_backbone import FalconBackbone
from keras_hub.src.models.falcon.falcon_causal_lm import FalconCausalLM
from keras_hub.src.models.falcon.falcon_causal_lm_preprocessor import (
    FalconCausalLMPreprocessor,
)
from keras_hub.src.models.falcon.falcon_tokenizer import FalconTokenizer
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone
from keras_hub.src.models.flux.flux_model import FluxBackbone
from keras_hub.src.models.flux.flux_text_to_image import FluxTextToImage
from keras_hub.src.models.flux.flux_text_to_image_preprocessor import (
    FluxTextToImagePreprocessor,
)
from keras_hub.src.models.gemma.gemma_backbone import GemmaBackbone
from keras_hub.src.models.gemma.gemma_causal_lm import GemmaCausalLM
from keras_hub.src.models.gemma.gemma_causal_lm_preprocessor import (
    GemmaCausalLMPreprocessor,
)
from keras_hub.src.models.gemma.gemma_tokenizer import GemmaTokenizer
from keras_hub.src.models.gpt2.gpt2_backbone import GPT2Backbone
from keras_hub.src.models.gpt2.gpt2_causal_lm import GPT2CausalLM
from keras_hub.src.models.gpt2.gpt2_causal_lm_preprocessor import (
    GPT2CausalLMPreprocessor,
)
from keras_hub.src.models.gpt2.gpt2_preprocessor import GPT2Preprocessor
from keras_hub.src.models.gpt2.gpt2_tokenizer import GPT2Tokenizer
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_backbone import GPTNeoXBackbone
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_causal_lm import GPTNeoXCausalLM
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_causal_lm_preprocessor import (
    GPTNeoXCausalLMPreprocessor,
)
from keras_hub.src.models.gpt_neo_x.gpt_neo_x_tokenizer import GPTNeoXTokenizer
from keras_hub.src.models.image_classifier import ImageClassifier
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)
from keras_hub.src.models.image_segmenter import ImageSegmenter
from keras_hub.src.models.image_segmenter_preprocessor import (
    ImageSegmenterPreprocessor,
)
from keras_hub.src.models.image_to_image import ImageToImage
from keras_hub.src.models.inpaint import Inpaint
from keras_hub.src.models.llama.llama_backbone import LlamaBackbone
from keras_hub.src.models.llama.llama_causal_lm import LlamaCausalLM
from keras_hub.src.models.llama.llama_causal_lm_preprocessor import (
    LlamaCausalLMPreprocessor,
)
from keras_hub.src.models.llama.llama_tokenizer import LlamaTokenizer
from keras_hub.src.models.llama3.llama3_backbone import Llama3Backbone
from keras_hub.src.models.llama3.llama3_causal_lm import Llama3CausalLM
from keras_hub.src.models.llama3.llama3_causal_lm_preprocessor import (
    Llama3CausalLMPreprocessor,
)
from keras_hub.src.models.llama3.llama3_tokenizer import Llama3Tokenizer
from keras_hub.src.models.masked_lm import MaskedLM
from keras_hub.src.models.masked_lm_preprocessor import MaskedLMPreprocessor
from keras_hub.src.models.mistral.mistral_backbone import MistralBackbone
from keras_hub.src.models.mistral.mistral_causal_lm import MistralCausalLM
from keras_hub.src.models.mistral.mistral_causal_lm_preprocessor import (
    MistralCausalLMPreprocessor,
)
from keras_hub.src.models.mistral.mistral_tokenizer import MistralTokenizer
from keras_hub.src.models.mit.mit_backbone import MiTBackbone
from keras_hub.src.models.mit.mit_image_classifier import MiTImageClassifier
from keras_hub.src.models.mit.mit_image_classifier_preprocessor import (
    MiTImageClassifierPreprocessor,
)
from keras_hub.src.models.mobilenet.mobilenet_backbone import MobileNetBackbone
from keras_hub.src.models.mobilenet.mobilenet_image_classifier import (
    MobileNetImageClassifier,
)
from keras_hub.src.models.mobilenet.mobilenet_image_classifier_preprocessor import (
    MobileNetImageClassifierPreprocessor,
)
from keras_hub.src.models.object_detector import ObjectDetector
from keras_hub.src.models.object_detector import (
    ObjectDetector as ImageObjectDetector,
)
from keras_hub.src.models.object_detector_preprocessor import (
    ObjectDetectorPreprocessor,
)
from keras_hub.src.models.object_detector_preprocessor import (
    ObjectDetectorPreprocessor as ImageObjectDetectorPreprocessor,
)
from keras_hub.src.models.opt.opt_backbone import OPTBackbone
from keras_hub.src.models.opt.opt_causal_lm import OPTCausalLM
from keras_hub.src.models.opt.opt_causal_lm_preprocessor import (
    OPTCausalLMPreprocessor,
)
from keras_hub.src.models.opt.opt_tokenizer import OPTTokenizer
from keras_hub.src.models.pali_gemma.pali_gemma_backbone import (
    PaliGemmaBackbone,
)
from keras_hub.src.models.pali_gemma.pali_gemma_causal_lm import (
    PaliGemmaCausalLM,
)
from keras_hub.src.models.pali_gemma.pali_gemma_causal_lm_preprocessor import (
    PaliGemmaCausalLMPreprocessor,
)
from keras_hub.src.models.pali_gemma.pali_gemma_tokenizer import (
    PaliGemmaTokenizer,
)
from keras_hub.src.models.phi3.phi3_backbone import Phi3Backbone
from keras_hub.src.models.phi3.phi3_causal_lm import Phi3CausalLM
from keras_hub.src.models.phi3.phi3_causal_lm_preprocessor import (
    Phi3CausalLMPreprocessor,
)
from keras_hub.src.models.phi3.phi3_tokenizer import Phi3Tokenizer
from keras_hub.src.models.preprocessor import Preprocessor
from keras_hub.src.models.resnet.resnet_backbone import ResNetBackbone
from keras_hub.src.models.resnet.resnet_image_classifier import (
    ResNetImageClassifier,
)
from keras_hub.src.models.resnet.resnet_image_classifier_preprocessor import (
    ResNetImageClassifierPreprocessor,
)
from keras_hub.src.models.retinanet.retinanet_backbone import RetinaNetBackbone
from keras_hub.src.models.retinanet.retinanet_object_detector import (
    RetinaNetObjectDetector,
)
from keras_hub.src.models.retinanet.retinanet_object_detector_preprocessor import (
    RetinaNetObjectDetectorPreprocessor,
)
from keras_hub.src.models.roberta.roberta_backbone import RobertaBackbone
from keras_hub.src.models.roberta.roberta_masked_lm import RobertaMaskedLM
from keras_hub.src.models.roberta.roberta_masked_lm_preprocessor import (
    RobertaMaskedLMPreprocessor,
)
from keras_hub.src.models.roberta.roberta_text_classifier import (
    RobertaTextClassifier,
)
from keras_hub.src.models.roberta.roberta_text_classifier import (
    RobertaTextClassifier as RobertaClassifier,
)
from keras_hub.src.models.roberta.roberta_text_classifier_preprocessor import (
    RobertaTextClassifierPreprocessor,
)
from keras_hub.src.models.roberta.roberta_text_classifier_preprocessor import (
    RobertaTextClassifierPreprocessor as RobertaPreprocessor,
)
from keras_hub.src.models.roberta.roberta_tokenizer import RobertaTokenizer
from keras_hub.src.models.sam.sam_backbone import SAMBackbone
from keras_hub.src.models.sam.sam_image_segmenter import SAMImageSegmenter
from keras_hub.src.models.sam.sam_image_segmenter_preprocessor import (
    SAMImageSegmenterPreprocessor,
)
from keras_hub.src.models.segformer.segformer_backbone import SegFormerBackbone
from keras_hub.src.models.segformer.segformer_image_segmenter import (
    SegFormerImageSegmenter,
)
from keras_hub.src.models.segformer.segformer_image_segmenter_preprocessor import (
    SegFormerImageSegmenterPreprocessor,
)
from keras_hub.src.models.seq_2_seq_lm import Seq2SeqLM
from keras_hub.src.models.seq_2_seq_lm_preprocessor import Seq2SeqLMPreprocessor
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_backbone import (
    StableDiffusion3Backbone,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_image_to_image import (
    StableDiffusion3ImageToImage,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_inpaint import (
    StableDiffusion3Inpaint,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_text_to_image import (
    StableDiffusion3TextToImage,
)
from keras_hub.src.models.stable_diffusion_3.stable_diffusion_3_text_to_image_preprocessor import (
    StableDiffusion3TextToImagePreprocessor,
)
from keras_hub.src.models.t5.t5_backbone import T5Backbone
from keras_hub.src.models.t5.t5_preprocessor import T5Preprocessor
from keras_hub.src.models.t5.t5_tokenizer import T5Tokenizer
from keras_hub.src.models.task import Task
from keras_hub.src.models.text_classifier import TextClassifier
from keras_hub.src.models.text_classifier import TextClassifier as Classifier
from keras_hub.src.models.text_classifier_preprocessor import (
    TextClassifierPreprocessor,
)
from keras_hub.src.models.text_to_image import TextToImage
from keras_hub.src.models.vgg.vgg_backbone import VGGBackbone
from keras_hub.src.models.vgg.vgg_image_classifier import VGGImageClassifier
from keras_hub.src.models.vgg.vgg_image_classifier_preprocessor import (
    VGGImageClassifierPreprocessor,
)
from keras_hub.src.models.vit.vit_backbone import ViTBackbone
from keras_hub.src.models.vit.vit_image_classifier import ViTImageClassifier
from keras_hub.src.models.vit.vit_image_classifier_preprocessor import (
    ViTImageClassifierPreprocessor,
)
from keras_hub.src.models.vit_det.vit_det_backbone import ViTDetBackbone
from keras_hub.src.models.whisper.whisper_backbone import WhisperBackbone
from keras_hub.src.models.whisper.whisper_tokenizer import WhisperTokenizer
from keras_hub.src.models.xlm_roberta.xlm_roberta_backbone import (
    XLMRobertaBackbone,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_masked_lm import (
    XLMRobertaMaskedLM,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_masked_lm_preprocessor import (
    XLMRobertaMaskedLMPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier import (
    XLMRobertaTextClassifier,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier import (
    XLMRobertaTextClassifier as XLMRobertaClassifier,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier_preprocessor import (
    XLMRobertaTextClassifierPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_text_classifier_preprocessor import (
    XLMRobertaTextClassifierPreprocessor as XLMRobertaPreprocessor,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_tokenizer import (
    XLMRobertaTokenizer,
)
from keras_hub.src.models.xlnet.xlnet_backbone import XLNetBackbone
from keras_hub.src.tokenizers.tokenizer import Tokenizer



import numpy as np

from keras_hub.src.models.vit.vit_backbone import ViTBackbone

backbone_cls = ViTBackbone


def convert_backbone_config(transformers_config):
    image_size = transformers_config["image_size"]
    return {
        "image_shape": (image_size, image_size, 3),
        "patch_size": transformers_config["patch_size"],
        "num_layers": transformers_config["num_hidden_layers"],
        "num_heads": transformers_config["num_attention_heads"],
        "hidden_dim": transformers_config["hidden_size"],
        "mlp_dim": transformers_config["intermediate_size"],
        "dropout_rate": transformers_config["hidden_dropout_prob"],
        "attention_dropout": transformers_config[
            "attention_probs_dropout_prob"
        ],
        "use_mha_bias": transformers_config["qkv_bias"],
    }


def convert_weights(backbone, loader, transformers_config):
    def port_ln(keras_variable, weight_key):
        loader.port_weight(keras_variable.gamma, f"{weight_key}.weight")
        loader.port_weight(keras_variable.beta, f"{weight_key}.bias")

    def port_dense(keras_variable, weight_key):
        loader.port_weight(
            keras_variable.kernel,
            f"{weight_key}.weight",
            hook_fn=lambda x, _: x.T,
        )
        if keras_variable.bias is not None:
            loader.port_weight(keras_variable.bias, f"{weight_key}.bias")

    def port_mha(keras_variable, weight_key, num_heads, hidden_dim):
        # query
        loader.port_weight(
            keras_variable.query_dense.kernel,
            f"{weight_key}.attention.query.weight",
            hook_fn=lambda x, _: np.reshape(
                x.T, (hidden_dim, num_heads, hidden_dim // num_heads)
            ),
        )
        loader.port_weight(
            keras_variable.query_dense.bias,
            f"{weight_key}.attention.query.bias",
            hook_fn=lambda x, _: np.reshape(
                x, (num_heads, hidden_dim // num_heads)
            ),
        )
        # key
        loader.port_weight(
            keras_variable.key_dense.kernel,
            f"{weight_key}.attention.key.weight",
            hook_fn=lambda x, _: np.reshape(
                x.T, (hidden_dim, num_heads, hidden_dim // num_heads)
            ),
        )
        loader.port_weight(
            keras_variable.key_dense.bias,
            f"{weight_key}.attention.key.bias",
            hook_fn=lambda x, _: np.reshape(
                x, (num_heads, hidden_dim // num_heads)
            ),
        )
        # value
        loader.port_weight(
            keras_variable.value_dense.kernel,
            f"{weight_key}.attention.value.weight",
            hook_fn=lambda x, _: np.reshape(
                x.T, (hidden_dim, num_heads, hidden_dim // num_heads)
            ),
        )
        loader.port_weight(
            keras_variable.value_dense.bias,
            f"{weight_key}.attention.value.bias",
            hook_fn=lambda x, _: np.reshape(
                x, (num_heads, hidden_dim // num_heads)
            ),
        )
        # output
        loader.port_weight(
            keras_variable.output_dense.kernel,
            f"{weight_key}.output.dense.weight",
            hook_fn=lambda x, _: np.reshape(
                x.T, (num_heads, hidden_dim // num_heads, hidden_dim)
            ),
        )
        loader.port_weight(
            keras_variable.output_dense.bias, f"{weight_key}.output.dense.bias"
        )

    loader.port_weight(
        keras_variable=backbone.layers[1].patch_embedding.kernel,
        hf_weight_key="vit.embeddings.patch_embeddings.projection.weight",
        hook_fn=lambda x, _: np.transpose(x, (2, 3, 1, 0)),
    )

    loader.port_weight(
        backbone.layers[1].patch_embedding.bias,
        "vit.embeddings.patch_embeddings.projection.bias",
    )

    loader.port_weight(
        backbone.layers[1].class_token,
        "vit.embeddings.cls_token",
    )

    loader.port_weight(
        backbone.layers[1].position_embedding.embeddings,
        "vit.embeddings.position_embeddings",
        hook_fn=lambda x, _: x[0],
    )
    encoder_layers = backbone.layers[2].encoder_layers
    for i, encoder_block in enumerate(encoder_layers):
        prefix = "vit.encoder.layer"
        num_heads = encoder_block.num_heads
        hidden_dim = encoder_block.hidden_dim

        port_mha(
            encoder_block.mha,
            f"{prefix}.{i}.attention",
            num_heads,
            hidden_dim,
        )
        port_ln(encoder_block.layer_norm_1, f"{prefix}.{i}.layernorm_before")
        port_ln(encoder_block.layer_norm_2, f"{prefix}.{i}.layernorm_after")

        port_dense(
            encoder_block.mlp.dense_1, f"{prefix}.{i}.intermediate.dense"
        )
        port_dense(encoder_block.mlp.dense_2, f"{prefix}.{i}.output.dense")
    port_ln(backbone.layers[2].layer_norm, "vit.layernorm")


def convert_head(task, loader, transformers_config):
    prefix = "classifier."
    loader.port_weight(
        task.output_dense.kernel,
        hf_weight_key=prefix + "weight",
        hook_fn=lambda x, _: x.T,
    )
    loader.port_weight(
        task.output_dense.bias,
        hf_weight_key=prefix + "bias",
    )



from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.samplers.sampler import Sampler


@keras_hub_export("keras_hub.samplers.GreedySampler")
class GreedySampler(Sampler):
    """Greedy sampler class.

    This sampler is implemented on greedy search, i.e., always picking up the
    token of the largest probability as the next token.

    Examples:
    ```python
    causal_lm = keras_hub.models.GPT2CausalLM.from_preset("gpt2_base_en")

    # Pass by name to compile.
    causal_lm.compile(sampler="greedy")
    causal_lm.generate(["Keras is a"])

    # Pass by object to compile.
    sampler = keras_hub.samplers.GreedySampler()
    causal_lm.compile(sampler=sampler)
    causal_lm.generate(["Keras is a"])
    ```
    """

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def get_next_token(self, probabilities):
        return ops.argmax(probabilities, axis=-1)



"""SAM preset configurations."""

backbone_presets = {
    "sam_base_sa1b": {
        "metadata": {
            "description": ("The base SAM model trained on the SA1B dataset."),
            "params": 93735728,
            "path": "sam",
        },
        "kaggle_handle": "kaggle://keras/sam/keras/sam_base_sa1b/5",
    },
    "sam_large_sa1b": {
        "metadata": {
            "description": ("The large SAM model trained on the SA1B dataset."),
            "params": 641090864,
            "path": "sam",
        },
        "kaggle_handle": "kaggle://keras/sam/keras/sam_large_sa1b/5",
    },
    "sam_huge_sa1b": {
        "metadata": {
            "description": ("The huge SAM model trained on the SA1B dataset."),
            "params": 312343088,
            "path": "sam",
        },
        "kaggle_handle": "kaggle://keras/sam/keras/sam_huge_sa1b/5",
    },
}



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.start_end_packer import StartEndPacker
from keras_hub.src.models.bart.bart_backbone import BartBackbone
from keras_hub.src.models.bart.bart_tokenizer import BartTokenizer
from keras_hub.src.models.seq_2_seq_lm_preprocessor import Seq2SeqLMPreprocessor


@keras_hub_export("keras_hub.models.BartSeq2SeqLMPreprocessor")
class BartSeq2SeqLMPreprocessor(Seq2SeqLMPreprocessor):
    """BART Seq2Seq LM preprocessor.

    This layer is used as preprocessor for seq2seq tasks using the BART model.
    This class subclasses `keras_hub.models.BartPreprocessor` and keeps most of
    its functionality. It has two changes from the superclass:

     1. Sets the `y` (label) and `sample_weights` fields by shifting the
        decoder input sequence one step towards the left. Both these fields are
        inferred internally, and any passed values will be ignored.
     2. Drops the last token from the decoder input sequence as it does not have
        a successor.

    Args:
        tokenizer: A `keras_hub.models.BartTokenizer` instance.
        encoder_sequence_length: The length of the packed encoder inputs.
        decoder_sequence_length: The length of the packed decoder inputs.

    Call arguments:
        x: A dictionary with `encoder_text` and `decoder_text` as its keys.
            Each value in the dictionary should be a tensor of single string
            sequences. Inputs may be batched or unbatched. Raw python inputs
            will be converted to tensors.
        y: Label data. Should always be `None` as the layer generates labels by
            shifting the decoder input sequence one step to the left.
        sample_weight: Label weights. Should always be `None` as the layer
            generates label weights by shifting the padding mask one step to the
            left.

    Examples:

    Directly calling the layer on data
    ```python
    preprocessor = keras_hub.models.BartPreprocessor.from_preset("bart_base_en")

    # Preprocess unbatched inputs.
    inputs = {
        "encoder_text": "The fox was sleeping.",
        "decoder_text": "The fox was awake."
    }
    preprocessor(inputs)

    # Preprocess batched inputs.
    inputs = {
        "encoder_text": ["The fox was sleeping.", "The lion was quiet."],
        "decoder_text": ["The fox was awake.", "The lion was roaring."]
    }
    preprocessor(inputs)

    # Custom vocabulary.
    vocab = {
        "<s>": 0,
        "<pad>": 1,
        "</s>": 2,
        "Ġafter": 5,
        "noon": 6,
        "Ġsun": 7,
    }
    merges = ["Ġ a", "Ġ s", "Ġ n", "e r", "n o", "o n", "Ġs u", "Ġa f", "no on"]
    merges += ["Ġsu n", "Ġaf t", "Ġaft er"]

    tokenizer = keras_hub.models.BartTokenizer(
        vocabulary=vocab,
        merges=merges,
    )
    preprocessor = keras_hub.models.BartPreprocessor(
        tokenizer=tokenizer,
        encoder_sequence_length=20,
        decoder_sequence_length=10,
    )
    inputs = {
        "encoder_text": "The fox was sleeping.",
        "decoder_text": "The fox was awake."
    }
    preprocessor(inputs)
    ```

    Mapping with `tf.data.Dataset`.
    ```python
    preprocessor = keras_hub.models.BartPreprocessor.from_preset("bart_base_en")

    # Map single sentences.
    features = {
        "encoder_text": tf.constant(
            ["The fox was sleeping.", "The lion was quiet."]
        ),
        "decoder_text": tf.constant(
            ["The fox was awake.", "The lion was roaring."]
        )
    }
    ds = tf.data.Dataset.from_tensor_slices(features)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)
    ```
    """

    backbone_cls = BartBackbone
    tokenizer_cls = BartTokenizer

    def build(self, input_shape):
        super().build(input_shape)
        # The decoder is packed a bit differently; the format is as follows:
        # `[end_token_id, start_token_id, tokens..., end_token_id, padding...]`.
        self.decoder_packer = StartEndPacker(
            start_value=[
                self.tokenizer.end_token_id,
                self.tokenizer.start_token_id,
            ],
            end_value=self.tokenizer.end_token_id,
            pad_value=self.tokenizer.pad_token_id,
            sequence_length=self.decoder_sequence_length,
            return_padding_mask=True,
        )



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone


@keras_hub_export("keras_hub.models.FeaturePyramidBackbone")
class FeaturePyramidBackbone(Backbone):
    """A backbone with feature pyramid outputs.

    `FeaturePyramidBackbone` extends `Backbone` with a single `pyramid_outputs`
    property for accessing the feature pyramid outputs of the model. Subclassers
    should set the `pyramid_outputs` property during the model constructor.

    Example:

    ```python
    input_data = np.random.uniform(0, 256, size=(2, 224, 224, 3))

    # Convert to feature pyramid output format using ResNet.
    backbone = ResNetBackbone.from_preset("resnet50")
    model = keras.Model(
        inputs=backbone.inputs, outputs=backbone.pyramid_outputs
    )
    model(input_data)  # A dict containing the keys ["P2", "P3", "P4", "P5"]
    ```
    """

    @property
    def pyramid_outputs(self):
        """A dict for feature pyramid outputs.

        The key is a string represents the name of the feature output and the
        value is a `keras.KerasTensor`. A typical feature pyramid has multiple
        levels corresponding to scales such as `["P2", "P3", "P4", "P5"]`. Scale
        `Pn` represents a feature map `2^n` times smaller in width and height
        than the inputs.
        """
        return getattr(self, "_pyramid_outputs", {})

    @pyramid_outputs.setter
    def pyramid_outputs(self, value):
        if not isinstance(value, dict):
            raise TypeError(
                "`pyramid_outputs` must be a dictionary. "
                f"Received: value={value} of type {type(value)}"
            )
        for k, v in value.items():
            if not isinstance(k, str):
                raise TypeError(
                    "The key of `pyramid_outputs` must be a string. "
                    f"Received: key={k} of type {type(k)}"
                )
            if not isinstance(v, keras.KerasTensor):
                raise TypeError(
                    "The value of `pyramid_outputs` must be a "
                    "`keras.KerasTensor`. "
                    f"Received: value={v} of type {type(v)}"
                )
        self._pyramid_outputs = value



"""BLOOM model preset configurations."""

backbone_presets = {
    "bloom_560m_multi": {
        "metadata": {
            "description": (
                "24-layer Bloom model with hidden dimension of 1024. "
                "trained on 45 natural languages and 12 programming languages."
            ),
            "params": 559214592,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloom_560m_multi/4",
    },
    "bloom_1.1b_multi": {
        "metadata": {
            "description": (
                "24-layer Bloom model with hidden dimension of 1536. "
                "trained on 45 natural languages and 12 programming languages."
            ),
            "params": 1065314304,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloom_1.1b_multi/2",
    },
    "bloom_1.7b_multi": {
        "metadata": {
            "description": (
                "24-layer Bloom model with hidden dimension of 2048. "
                "trained on 45 natural languages and 12 programming languages."
            ),
            "params": 1722408960,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloom_1.7b_multi/2",
    },
    "bloom_3b_multi": {
        "metadata": {
            "description": (
                "30-layer Bloom model with hidden dimension of 2560. "
                "trained on 45 natural languages and 12 programming languages."
            ),
            "params": 3002557440,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloom_3b_multi/2",
    },
    "bloomz_560m_multi": {
        "metadata": {
            "description": (
                "24-layer Bloom model with hidden dimension of 1024. "
                "finetuned on crosslingual task mixture (xP3) dataset."
            ),
            "params": 559214592,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloomz_560m_multi/2",
    },
    "bloomz_1.1b_multi": {
        "metadata": {
            "description": (
                "24-layer Bloom model with hidden dimension of 1536. "
                "finetuned on crosslingual task mixture (xP3) dataset."
            ),
            "params": 1065314304,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloomz_1.1b_multi/2",
    },
    "bloomz_1.7b_multi": {
        "metadata": {
            "description": (
                "24-layer Bloom model with hidden dimension of 2048. "
                "finetuned on crosslingual task mixture (xP3) dataset."
            ),
            "params": 1722408960,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloomz_1.7b_multi/2",
    },
    "bloomz_3b_multi": {
        "metadata": {
            "description": (
                "30-layer Bloom model with hidden dimension of 2560. "
                "finetuned on crosslingual task mixture (xP3) dataset."
            ),
            "params": 3002557440,
            "path": "bloom",
        },
        "kaggle_handle": "kaggle://keras/bloom/keras/bloomz_3b_multi/2",
    },
}
